###########################################################################################
#  !!! Important
# Under development and testing: reducing memory footprint using boundary-saving methods
# [1] P. Yang, J. Gao, and B. Wang, RTM using effective boundary saving: A staggered grid GPU implementation, Comput. Geosci., vol. 68, pp. 64–72, Jul. 2014, doi: 10.1016/j.cageo.2014.04.004.
# [2] Wang, S., Jiang, Y., Song, P., Tan, J., Liu, Z. & He, B., 2023. Memory optimization in RNN-based full waveform inversion using boundary saving wavefield reconstruction. IEEE Trans. Geosci. Remote Sens., 61, 1–12. doi:10.1109/TGRS.2023.3317529
###########################################################################################
import torch
torch.autograd.set_detect_anomaly(True)
from torch import Tensor
from torch.utils.checkpoint import checkpoint
import numpy as np
from typing import Tuple,Dict
import itertools


def pad_torchSingle(v: torch.Tensor, pml: int, nz: int, nx: int, ns: int, device: torch.device = torch.device("cpu")) -> torch.Tensor:
    nz_pml = nz + 2 * pml
    nx_pml = nx + 2 * pml
    cc = torch.zeros((nz_pml, nx_pml), device=device)
    
    # Copy the original tensor to the appropriate position
    cc[pml:nz_pml - pml, pml:nx_pml - pml] = v

    # Handle the top boundary
    cc[:pml, pml:pml + nx] = cc[pml, pml:pml + nx].expand(pml, -1)
    
    # Handle the bottom boundary
    cc[nz_pml - pml:nz_pml, pml:pml + nx] = cc[nz_pml - pml - 1, pml:pml + nx].expand(pml, -1)

    # Handle the left boundary
    cc[:, :pml] = cc[:, [pml]].expand(-1, pml)

    # Handle the right boundary
    cc[:, nx_pml - pml:nx_pml] = cc[:, [nx_pml - pml - 1]].expand(-1, pml)

    return cc

# 2Nth order staggered grid finite difference : correct backward propagation needs 2N -1 points on one side
# the default is N = 3 (2N-1)
def save_boundaries(tensor: torch.Tensor, nabc: int, N: int=3) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
    top     =  tensor[:, nabc:nabc + N, :].clone()
    bottom  =  tensor[:, -(nabc + N):-nabc, :].clone()
    left    = tensor[..., nabc:nabc + N].clone() 
    right   = tensor[..., -(nabc + N):-nabc].clone()
    return top, bottom, left, right

def restore_boundaries(tensor: torch.Tensor, memory: Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor], nabc: int, N: int = 3) -> torch.Tensor:
    top, bottom, left, right        = memory
    # Restore boundaries from memory
    tensor[:, nabc:nabc + N, :]     = top
    tensor[:, -(nabc + N):-nabc, :] = bottom
    tensor[..., nabc:nabc + N]      = left
    tensor[..., -(nabc + N):-nabc]  = right
    return tensor

def _time_step(
        src_n: int, src_x: torch.Tensor, src_z: torch.Tensor,src_v: torch.Tensor, # source information
        nx_pml: int, nz_pml: int, nabc: int,                                                   
        free_surface: bool, free_surface_start: int,                    
        kappa1: torch.Tensor, alpha1: torch.Tensor, 
        kappa2: torch.Tensor, alpha2: torch.Tensor, 
        kappa3: torch.Tensor,  
        p: torch.Tensor, u: torch.Tensor, w: torch.Tensor,
        ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:

    # Coefficients for the staggered grid
    c1_staggered = 9.0 / 8.0
    c2_staggered = -1.0 / 24.0
    
    # Update pressure
    p[:, free_surface_start + 1:nz_pml - 2, 2:nx_pml - 2] = (
        (1.0 - kappa1[free_surface_start + 1:nz_pml - 2, 2:nx_pml - 2]) * 
        p[:, free_surface_start + 1:nz_pml - 2, 2:nx_pml - 2] - 
        alpha1[free_surface_start + 1:nz_pml - 2, 2:nx_pml - 2] * (
            c1_staggered * (u[:, free_surface_start + 1:nz_pml - 2, 2:nx_pml - 2] -
                            u[:, free_surface_start + 1:nz_pml - 2, 1:nx_pml - 3] +
                            w[:, free_surface_start + 1:nz_pml - 2, 2:nx_pml - 2] -
                            w[:, free_surface_start:nz_pml - 3, 2:nx_pml - 2]) +
            c2_staggered * (u[:, free_surface_start + 1:nz_pml - 2, 3:nx_pml - 1] -
                            u[:, free_surface_start + 1:nz_pml - 2, 0:nx_pml - 4] +
                            w[:, free_surface_start + 2:nz_pml - 1, 2:nx_pml - 2] -
                            w[:, free_surface_start - 1:nz_pml - 4, 2:nx_pml - 2])
        )
    )
    # Add source
    p[torch.arange(src_n), src_z, src_x] += src_v

    # Free surface handling
    if free_surface:
        p[:, free_surface_start - 1, :] = -p[:, free_surface_start + 1, :]

    # Update horizontal particle velocity: u
    u[:, free_surface_start:nz_pml - 1, 1:nx_pml - 2] = (
        (1.0 - kappa2[free_surface_start:nz_pml - 1, 1:nx_pml - 2]) * 
        u[:, free_surface_start:nz_pml - 1, 1:nx_pml - 2] - 
        alpha2[free_surface_start:nz_pml - 1, 1:nx_pml - 2] * (
            c1_staggered * (p[:, free_surface_start:nz_pml - 1, 2:nx_pml - 1] -
                            p[:, free_surface_start:nz_pml - 1, 1:nx_pml - 2]) +
            c2_staggered * (p[:, free_surface_start:nz_pml - 1, 3:nx_pml] -
                            p[:, free_surface_start:nz_pml - 1, 0:nx_pml - 3])
        )
    )

    # Update vertical particle velocity: w
    w[:, free_surface_start:nz_pml - 2, 1:nx_pml - 1] = (
        (1.0 - kappa3[free_surface_start:nz_pml - 2, 1:nx_pml - 1]) *
        w[:, free_surface_start:nz_pml - 2, 1:nx_pml - 1] - 
        alpha2[free_surface_start:nz_pml - 2, 1:nx_pml - 1] * (
            c1_staggered * (p[:, free_surface_start + 1:nz_pml - 1, 1:nx_pml - 1] -
                            p[:, free_surface_start:nz_pml - 2, 1:nx_pml - 1]) +
            c2_staggered * (p[:, free_surface_start + 2:nz_pml, 1:nx_pml - 1] -
                            p[:, free_surface_start - 1:nz_pml - 3, 1:nx_pml - 1])
        )
    )

    # Free surface for vertical velocity
    if free_surface:
        w[:, free_surface_start - 1, :] = w[:, free_surface_start, :]
        
    return p, u, w

def _time_step_backward(
        src_n: int, src_x: torch.Tensor, src_z: torch.Tensor,src_v: torch.Tensor, # source information
        nx_pml: int, nz_pml: int, nabc: int,                                                    
        free_surface: bool, free_surface_start: int,                    
        kappa1: torch.Tensor, alpha1: torch.Tensor, 
        kappa2: torch.Tensor, alpha2: torch.Tensor, 
        kappa3: torch.Tensor,  
        p_last: torch.Tensor, u_last: torch.Tensor, w_last: torch.Tensor,
        p_boundary_memory: torch.Tensor,u_boundary_memory: torch.Tensor,w_boundary_memory: torch.Tensor
        ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:

    p = p_last.clone()
    u = u_last.clone()
    w = w_last.clone()
    
    # Coefficients for the staggered grid
    c1_staggered = 9.0 / 8.0
    c2_staggered = -1.0 / 24.0
    
    # Update pressure
    p[:, free_surface_start + 1:nz_pml - 2, 2:nx_pml - 2] = (
        (1.0 - kappa1[free_surface_start + 1:nz_pml - 2, 2:nx_pml - 2]) * 
        p[:, free_surface_start + 1:nz_pml - 2, 2:nx_pml - 2] - 
        alpha1[free_surface_start + 1:nz_pml - 2, 2:nx_pml - 2] * (
            c1_staggered * (u[:, free_surface_start + 1:nz_pml - 2, 2:nx_pml - 2] -
                            u[:, free_surface_start + 1:nz_pml - 2, 1:nx_pml - 3] +
                            w[:, free_surface_start + 1:nz_pml - 2, 2:nx_pml - 2] -
                            w[:, free_surface_start:nz_pml - 3, 2:nx_pml - 2]) +
            c2_staggered * (u[:, free_surface_start + 1:nz_pml - 2, 3:nx_pml - 1] -
                            u[:, free_surface_start + 1:nz_pml - 2, 0:nx_pml - 4] +
                            w[:, free_surface_start + 2:nz_pml - 1, 2:nx_pml - 2] -
                            w[:, free_surface_start - 1:nz_pml - 4, 2:nx_pml - 2])
        )
    )
    
    # Add source
    p[torch.arange(src_n), src_z, src_x] += src_v

    # Free surface handling
    if free_surface:
        p[:, free_surface_start - 1, :] = -p[:, free_surface_start + 1, :]

    # Update horizontal particle velocity: u
    u[:, free_surface_start:nz_pml - 1, 1:nx_pml - 2] = (
        (1.0 - kappa2[free_surface_start:nz_pml - 1, 1:nx_pml - 2]) * 
        u[:, free_surface_start:nz_pml - 1, 1:nx_pml - 2] - 
        alpha2[free_surface_start:nz_pml - 1, 1:nx_pml - 2] * (
            c1_staggered * (p[:, free_surface_start:nz_pml - 1, 2:nx_pml - 1] -
                            p[:, free_surface_start:nz_pml - 1, 1:nx_pml - 2]) +
            c2_staggered * (p[:, free_surface_start:nz_pml - 1, 3:nx_pml] -
                            p[:, free_surface_start:nz_pml - 1, 0:nx_pml - 3])
        )
    )

    # Update vertical particle velocity: w
    w[:, free_surface_start:nz_pml - 2, 1:nx_pml - 1] = (
        (1.0 - kappa3[free_surface_start:nz_pml - 2, 1:nx_pml - 1]) *
        w[:, free_surface_start:nz_pml - 2, 1:nx_pml - 1] - 
        alpha2[free_surface_start:nz_pml - 2, 1:nx_pml - 1] * (
            c1_staggered * (p[:, free_surface_start + 1:nz_pml - 1, 1:nx_pml - 1] -
                            p[:, free_surface_start:nz_pml - 2, 1:nx_pml - 1]) +
            c2_staggered * (p[:, free_surface_start + 2:nz_pml, 1:nx_pml - 1] -
                            p[:, free_surface_start - 1:nz_pml - 3, 1:nx_pml - 1])
        )
    )

    # Free surface for vertical velocity
    if free_surface:
        w[:, free_surface_start - 1, :] = w[:, free_surface_start, :]

    with torch.no_grad():
        p = restore_boundaries(p, p_boundary_memory,nabc)
        u = restore_boundaries(u, u_boundary_memory,nabc)
        w = restore_boundaries(w, w_boundary_memory,nabc)

    return p, u, w

def packup_boundaries(flattened_list, tuple_size):
    return [tuple(flattened_list[i:i+tuple_size]) for i in range(0, len(flattened_list), tuple_size)]

class Checkpoint_TimeStep(torch.autograd.Function):
    wavefields = []
    counts=0
    @staticmethod
    def forward(ctx, 
                run_function,   # time_step forward
                back_function,  # time_step backward
                save_condition, # save the wavefield or not
                src_n: int, src_x: torch.Tensor, src_z: torch.Tensor,src_v: torch.Tensor, # source information
                nx_pml: int, nz_pml: int, nabc:int,                                                  
                free_surface: bool, free_surface_start: int,                    
                kappa1: torch.Tensor, alpha1: torch.Tensor, 
                kappa2: torch.Tensor, alpha2: torch.Tensor, 
                kappa3: torch.Tensor,  
                p: torch.Tensor, u: torch.Tensor, w: torch.Tensor
    ):  
        # require gradients or not
        requires_grad_list      = [False]*17
        for i,val in enumerate([kappa1,alpha1,kappa2,alpha2,kappa3,p,u,w]):
            if val.requires_grad:
                requires_grad_list[i+9] = True
        ctx.requires_grad_list  = requires_grad_list
        
        # forward simulation
        ctx.run_function        = run_function
        ctx.back_function       = back_function
        
        # save the forward parameters
        ctx.inputs_param = (src_n, src_x, src_z, src_v,
                            nx_pml, nz_pml, nabc,
                            free_surface, free_surface_start, 
                            kappa1, alpha1, kappa2, alpha2, kappa3
                            )
        
        # save the last image    
        ctx.save_condition      = save_condition
        
        with torch.no_grad():
            outputs = run_function(
                src_n, src_x, src_z, src_v,
                nx_pml, nz_pml, nabc,
                free_surface, free_surface_start, 
                kappa1, alpha1, kappa2, alpha2, kappa3, 
                p, u, w
            )
        
        # save boundary CheckpointFunction._bound.append(boundarys)
        boundarys = [save_boundaries(output,nabc) for output in outputs]
        ctx.save_for_backward(*itertools.chain(*boundarys))
        
        # Save the wavefields of the last time step
        ctx.is_last_time_step   = save_condition
        ctx.lastframe           = outputs if ctx.is_last_time_step else None
        return outputs
    
    @staticmethod
    def backward(ctx, *args):
        # the last wavefield
        if ctx.is_last_time_step:
            Checkpoint_TimeStep.wavefields = list(ctx.lastframe) # the last wavefields
            # Checkpoint_TimeStep.wavefields.reverse()             # reverse the wavefields
            wavefields = Checkpoint_TimeStep.wavefields
            return (None,None,None) + tuple(None for _ in range(len(ctx.requires_grad_list)))
        else:
            wavefields = Checkpoint_TimeStep.wavefields # get the wavefields

        Checkpoint_TimeStep.counts+=1
        
        #############################################################################
        # Inputs for backwards
        inputs = ctx.inputs_param + tuple(wavefields)
        # Process inputs to manage gradient requirements
        inputs_new = [
            inp.detach().requires_grad_(ctx.requires_grad_list[i]) if torch.is_tensor(inp) else inp
            for i, inp in enumerate(inputs)
        ]
        num_boundaries  = 4 
        boundaries      = packup_boundaries(ctx.saved_tensors, num_boundaries)
        # Combine new inputs with boundaries
        inputs = tuple(inputs_new) + tuple(boundaries)

        # reconstruct the wavefield
        with torch.enable_grad():
            outputs = ctx.back_function(*inputs)

        # calculate the gradients
        outputs_with_grad = []
        args_with_grad    = []
        for i in range(len(outputs)):
            if torch.is_tensor(outputs[i]) and outputs[i].requires_grad:
                outputs_with_grad.append(outputs[i])
                args_with_grad.append(args[i])
        
        if len(outputs_with_grad) == 0:
            raise RuntimeError(
                "none of output has requires_grad=True,"
                " this checkpoint() is not necessary")

        torch.autograd.backward(outputs_with_grad, args_with_grad)
        
        # assign boundary values
        outputs = list(outputs)
        
        # Update wavefields
        if not (Checkpoint_TimeStep.counts == 1) or not Checkpoint_TimeStep.counts == 0:
            Checkpoint_TimeStep.wavefields.clear()
            Checkpoint_TimeStep.wavefields.extend(list(outputs))
        
        grads = (None,None,None) + tuple(inp.grad if isinstance(inp, torch.Tensor) else None
                      for inp in inputs[:len(ctx.requires_grad_list)])
        
        return grads


def step_forward(nx: int, nz: int, dx: float, dz: float, dt: float,
                 nabc: int, free_surface: bool,                                             # Model settings
                 src_x: torch.Tensor, src_z: torch.Tensor, src_n: int, src_v: torch.Tensor, # Source
                 rcv_x: torch.Tensor, rcv_z: torch.Tensor, rcv_n: int,                      # Receiver
                 damp: torch.Tensor, v: torch.Tensor, rho: torch.Tensor,                    # Pass in velocity and density to compute alpha, kappa
                 p: torch.Tensor, u: torch.Tensor, w: torch.Tensor,
                 device: torch.device = torch.device("cpu"), dtype: torch.dtype = torch.float32) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
    
    nt = src_v.shape[-1]
    free_surface_start = nabc if free_surface else 1
    nx_pml = nx + 2 * nabc
    nz_pml = nz + 2 * nabc

    # Initialize recorded values
    rcv_p               = torch.zeros((src_n, nt, rcv_n), dtype=dtype, device=device)
    rcv_u               = torch.zeros((src_n, nt, rcv_n), dtype=dtype, device=device)
    rcv_w               = torch.zeros((src_n, nt, rcv_n), dtype=dtype, device=device)

    # Initialize forward wavefield
    forward_wavefield_p = torch.zeros((nz, nx), dtype=dtype, device=device)
    forward_wavefield_u = torch.zeros((nz, nx), dtype=dtype, device=device)
    forward_wavefield_w = torch.zeros((nz, nx), dtype=dtype, device=device)
    
    # Dynamically compute alpha and kappa at each time step to avoid storing large tensors
    alpha1 = rho * v * v * dt / dz
    kappa1 = damp * dt
    
    alpha2 = dt / (rho * dz)
    kappa2 = torch.zeros_like(damp, device=device)
    kappa2[:, 1:nx_pml - 2] = 0.5 * (damp[:, 1:nx_pml - 2] + damp[:, 2:nx_pml - 1]) * dt
    
    kappa3 = torch.zeros_like(damp, device=device)
    kappa3[free_surface_start:nz_pml - 2, :] = 0.5 * (damp[free_surface_start:nz_pml - 2, :] + damp[free_surface_start + 1:nz_pml - 1, :]) * dt

    time_step_fun = Checkpoint_TimeStep()
    for it in range(nt):
        # sources item
        src_v_it = (src_v[it]*dt if len(src_v.shape) == 1 else src_v[:, it])
        
        # Update fields and record values
        # p, u, w = _time_step(
        #     src_n, src_x, src_z, src_v_it,
        #     nx_pml, nz_pml, 
        #     free_surface, free_surface_start, 
        #     kappa1, alpha1, kappa2, alpha2, kappa3, 
        #     p, u, w
        # )
        # onlye the alpha1 requires gradients
        p,u,w = time_step_fun.apply(_time_step,_time_step_backward,it == nt-1,
                                    src_n, src_x, src_z, src_v_it,
                                    nx_pml, nz_pml, nabc,
                                    free_surface, free_surface_start, 
                                    kappa1, alpha1, kappa2, alpha2, kappa3, 
                                    p, u, w
                                )
        
        # record pressure seismogram
        rcv_p[:, it, :] = p[:, rcv_z, rcv_x]
        rcv_u[:, it, :] = u[:, rcv_z, rcv_x]
        rcv_w[:, it, :] = w[:, rcv_z, rcv_x]

        # Accumulate forward wavefields
        forward_wavefield_p += torch.sum(p * p, dim=0)[nabc:nabc + nz, nabc:nabc + nx].detach()
        forward_wavefield_u += torch.sum(u * u, dim=0)[nabc:nabc + nz, nabc:nabc + nx].detach()
        forward_wavefield_w += torch.sum(w * w, dim=0)[nabc:nabc + nz, nabc:nabc + nx].detach()

    return p, u, w, rcv_p, rcv_u, rcv_w, forward_wavefield_p, forward_wavefield_u, forward_wavefield_w


def forward_kernel(nx: int, nz: int, dx: float, dz: float, nt: int, dt: float,
                   nabc: int, free_surface: bool,                                                   # Model settings
                   src_x: torch.Tensor, src_z: torch.Tensor, src_n: int, src_v: torch.Tensor,       # Source
                   rcv_x: torch.Tensor, rcv_z: torch.Tensor, rcv_n: int,                            # Receiver
                   damp: torch.Tensor,                                                              # PML
                   v: torch.Tensor, rho: torch.Tensor,                                              # Velocity model
                   checkpoint_segments: int = 1,                                                    # Finite Difference
                   device: torch.device = torch.device('cpu'), dtype: torch.dtype = torch.float32
                   ) -> Dict[str, torch.Tensor]:  # Changed return type to Dict for clarity
    ###################################################################################
    # Padding velocity and density fields
    c   = pad_torchSingle(v, nabc, nz, nx, src_n, device=device)
    den = pad_torchSingle(rho, nabc, nz, nx, src_n, device=device)
    
    free_surface_start = nabc if free_surface else 1
    
    nx_pml = nx + 2 * nabc
    nz_pml = nz + 2 * nabc
    
    # Adjust source and receiver positions for PML
    src_x = src_x + nabc
    src_z = src_z + nabc
    rcv_x = rcv_x + nabc
    rcv_z = rcv_z + nabc
    
    # Initialize pressure, velocity fields
    p = torch.zeros((src_n, nz_pml, nx_pml)    , dtype=dtype, device=device)
    u = torch.zeros((src_n, nz_pml, nx_pml - 1), dtype=dtype, device=device)
    w = torch.zeros((src_n, nz_pml - 1, nx_pml), dtype=dtype, device=device)

    # Initialize recorded waveforms
    rcv_p = torch.zeros((src_n, nt, rcv_n), dtype=dtype, device=device)
    rcv_u = torch.zeros((src_n, nt, rcv_n), dtype=dtype, device=device)
    rcv_w = torch.zeros((src_n, nt, rcv_n), dtype=dtype, device=device)
    forward_wavefield_p = torch.zeros((nz, nx), dtype=dtype, device=device)
    forward_wavefield_u = torch.zeros((nz, nx), dtype=dtype, device=device)
    forward_wavefield_w = torch.zeros((nz, nx), dtype=dtype, device=device)

    k = 0
    for i, chunk in enumerate(torch.chunk(src_v, checkpoint_segments, dim=-1)):
        # Step forward: Calculate alpha, kappa inside step_forward
        p, u, w, rcv_p_temp, rcv_u_temp, rcv_w_temp, forward_wavefield_p_temp, forward_wavefield_u_temp, forward_wavefield_w_temp = \
            checkpoint(step_forward,
                       nx, nz, dx, dz, dt,
                       nabc, free_surface,
                       src_x, src_z, src_n, chunk,
                       rcv_x, rcv_z, rcv_n,
                       damp, c, den,  # Pass velocity, density, and damping (PML)
                       p, u, w,
                       device, dtype)

        # Save the waveform recorded on the receiver
        rcv_p[:, k:k + chunk.shape[-1]] = rcv_p_temp
        rcv_u[:, k:k + chunk.shape[-1]] = rcv_u_temp
        rcv_w[:, k:k + chunk.shape[-1]] = rcv_w_temp

        # Accumulate the forward wavefield
        forward_wavefield_p = forward_wavefield_p + forward_wavefield_p_temp.detach()
        forward_wavefield_u = forward_wavefield_p + forward_wavefield_u_temp.detach()
        forward_wavefield_w = forward_wavefield_p + forward_wavefield_w_temp.detach()
            
        k = k + chunk.shape[-1]
    
    record_waveform = {
        "p": rcv_p,
        "u": rcv_u,
        "w": rcv_w,
        "forward_wavefield_p": forward_wavefield_p,
        "forward_wavefield_u": forward_wavefield_u,
        "forward_wavefield_w": forward_wavefield_w,
    }
    
    return record_waveform



# def step_forward(nx: int, nz: int, dx: float, dz: float, dt: float,
#                  nabc: int, free_surface: bool,                                             # Model settings
#                  src_x: torch.Tensor, src_z: torch.Tensor, src_n: int, src_v: torch.Tensor, # Source
#                  rcv_x: torch.Tensor, rcv_z: torch.Tensor, rcv_n: int,                      # Receiver
#                  damp: torch.Tensor, v: torch.Tensor, rho: torch.Tensor,                    # Pass in velocity and density to compute alpha, kappa
#                  p: torch.Tensor, u: torch.Tensor, w: torch.Tensor,
#                  device: torch.device = torch.device("cpu"), dtype: torch.dtype = torch.float32) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
    
#     nt = src_v.shape[-1]
#     free_surface_start = nabc if free_surface else 1
#     nx_pml = nx + 2 * nabc
#     nz_pml = nz + 2 * nabc

#     # Initialize recorded values
#     rcv_p = torch.zeros((src_n, nt, rcv_n), dtype=dtype, device=device)
#     rcv_u = torch.zeros((src_n, nt, rcv_n), dtype=dtype, device=device)
#     rcv_w = torch.zeros((src_n, nt, rcv_n), dtype=dtype, device=device)

#     # Initialize forward wavefield
#     forward_wavefield_p = torch.zeros((nz, nx), dtype=dtype, device=device)
#     forward_wavefield_u = torch.zeros((nz, nx), dtype=dtype, device=device)
#     forward_wavefield_w = torch.zeros((nz, nx), dtype=dtype, device=device)
    
#     # Dynamically compute alpha and kappa at each time step to avoid storing large tensors
#     alpha1 = rho * v * v * dt / dz
#     kappa1 = damp * dt
    
#     alpha2 = dt / (rho * dz)
#     kappa2 = torch.zeros_like(damp, device=device)
#     kappa2[:, 1:nx_pml - 2] = 0.5 * (damp[:, 1:nx_pml - 2] + damp[:, 2:nx_pml - 1]) * dt
    
#     kappa3 = torch.zeros_like(damp, device=device)
#     kappa3[free_surface_start:nz_pml - 2, :] = 0.5 * (damp[free_surface_start:nz_pml - 2, :] + damp[free_surface_start + 1:nz_pml - 1, :]) * dt

#     for it in range(nt):
#         # sources item
#         src_v_it = (src_v[it]*dt if len(src_v.shape) == 1 else src_v[:, it])
        
#         # Update fields and record values
#         p, u, w = _time_step(
#             src_n, src_x, src_z, src_v_it,
#             nx_pml, nz_pml, 
#             free_surface, free_surface_start, 
#             kappa1, alpha1, kappa2, alpha2, kappa3, 
#             p, u, w
#         )
        
#         # record pressure seismogram
#         rcv_p[:, it, :] = p[:, rcv_z, rcv_x]
#         rcv_u[:, it, :] = u[:, rcv_z, rcv_x]
#         rcv_w[:, it, :] = w[:, rcv_z, rcv_x]

#         # Accumulate forward wavefields
#         forward_wavefield_p += torch.sum(p * p, dim=0)[nabc:nabc + nz, nabc:nabc + nx].detach()
#         forward_wavefield_u += torch.sum(u * u, dim=0)[nabc:nabc + nz, nabc:nabc + nx].detach()
#         forward_wavefield_w += torch.sum(w * w, dim=0)[nabc:nabc + nz, nabc:nabc + nx].detach()

#     return p, u, w, rcv_p, rcv_u, rcv_w, forward_wavefield_p, forward_wavefield_u, forward_wavefield_w




# def step_forward(nx: int, nz: int, dx: float, dz: float, dt: float,
#                  nabc: int, free_surface: bool,                               # Model settings
#                  src_x: torch.Tensor, src_z: torch.Tensor, src_n: int, src_v: torch.Tensor,     # Source
#                  rcv_x: torch.Tensor, rcv_z: torch.Tensor, rcv_n: int,                  # Receiver
#                  kappa1: torch.Tensor, alpha1: torch.Tensor, kappa2: torch.Tensor, alpha2: torch.Tensor,
#                  kappa3: torch.Tensor, c1_staggered: float, c2_staggered: float,
#                  p: torch.Tensor, u: torch.Tensor, w: torch.Tensor,
#                  device: torch.device = torch.device("cpu"), dtype: torch.dtype = torch.float32) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:

#     p = p.clone()
#     u = u.clone()
#     w = w.clone()
    
#     nt = src_v.shape[-1]
#     free_surface_start = nabc if free_surface else 1
#     nx_pml = nx + 2 * nabc
#     nz_pml = nz + 2 * nabc

#     # Initialize recorded values
#     rcv_p = torch.zeros((src_n, nt, rcv_n), dtype=dtype, device=device)
#     rcv_u = torch.zeros((src_n, nt, rcv_n), dtype=dtype, device=device)
#     rcv_w = torch.zeros((src_n, nt, rcv_n), dtype=dtype, device=device)

#     # Initialize forward wavefield
#     forward_wavefield_p = torch.zeros((nz, nx), dtype=dtype, device=device)
#     forward_wavefield_u = torch.zeros((nz, nx), dtype=dtype, device=device)
#     forward_wavefield_w = torch.zeros((nz, nx), dtype=dtype, device=device)
    
#     for it in range(nt):
#         # Update fields and record values
#         p, u, w = _time_step(
#             it, nx_pml, nz_pml, free_surface, free_surface_start, kappa1, alpha1, 
#             kappa2, alpha2, kappa3, c1_staggered, c2_staggered, dt, src_v, 
#             src_x, src_z, p, u, w, src_n, nabc, nx, nz
#         )
        
#         # Output pressure seismogram
#         rcv_p[:, it, :] = p[:, rcv_z, rcv_x]
#         rcv_u[:, it, :] = u[:, rcv_z, rcv_x]
#         rcv_w[:, it, :] = w[:, rcv_z, rcv_x]

#         # Accumulate forward wavefields
#         forward_wavefield_p += torch.sum(p * p, dim=0)[nabc:nabc + nz, nabc:nabc + nx].detach()
#         forward_wavefield_u += torch.sum(u * u, dim=0)[nabc:nabc + nz, nabc:nabc + nx].detach()
#         forward_wavefield_w += torch.sum(w * w, dim=0)[nabc:nabc + nz, nabc:nabc + nx].detach()

#     return p, u, w, rcv_p, rcv_u, rcv_w, forward_wavefield_p, forward_wavefield_u, forward_wavefield_w


# def forward_kernel(nx: int, nz: int, dx: float, dz: float, nt: int, dt: float,
#                    nabc: int, free_surface: bool,                               # Model settings
#                    src_x: torch.Tensor, src_z: torch.Tensor, src_n: int, src_v: torch.Tensor,     # Source
#                    rcv_x: torch.Tensor, rcv_z: torch.Tensor, rcv_n: int,                  # Receiver
#                    damp: torch.Tensor,                                              # PML
#                    v: torch.Tensor, rho: torch.Tensor,                                                    # Velocity model
#                    checkpoint_segments: int = 1,                                           # Finite Difference
#                    device: torch.device = torch.device('cpu'), dtype: torch.dtype = torch.float32
#                    ) -> Dict[str, torch.Tensor]:  # Changed return type to Dict for clarity
#     ###################################################################################
#     c   = pad_torchSingle(v, nabc, nz, nx, src_n, device=device)
#     den = pad_torchSingle(rho, nabc, nz, nx, src_n, device=device)
    
#     free_surface_start = nabc if free_surface else 1
    
#     nx_pml = nx + 2 * nabc
#     nz_pml = nz + 2 * nabc
    
#     src_x = src_x + nabc
#     src_z = src_z + nabc
    
#     rcv_x = rcv_x + nabc
#     rcv_z = rcv_z + nabc
    
#     # Initialize pressure, velocity fields
#     p = torch.zeros((src_n, nz_pml, nx_pml), dtype=dtype, device=device)
#     u = torch.zeros((src_n, nz_pml, nx_pml - 1), dtype=dtype, device=device)
#     w = torch.zeros((src_n, nz_pml - 1, nx_pml), dtype=dtype, device=device)

#     # Initialize recorded waveforms
#     rcv_p = torch.zeros((src_n, nt, rcv_n), dtype=dtype, device=device)
#     rcv_u = torch.zeros((src_n, nt, rcv_n), dtype=dtype, device=device)
#     rcv_w = torch.zeros((src_n, nt, rcv_n), dtype=dtype, device=device)
#     forward_wavefield_p = torch.zeros((nz, nx), dtype=dtype, device=device)
#     forward_wavefield_u = torch.zeros((nz, nx), dtype=dtype, device=device)
#     forward_wavefield_w = torch.zeros((nz, nx), dtype=dtype, device=device)

#     # Coefficients for the staggered grid
#     c1_staggered = 9.0 / 8.0
#     c2_staggered = -1.0 / 24.0
    
#     # Parameters for waveform simulation
#     alpha1 = den * c * c * dt / dz
#     kappa1 = damp * dt
    
#     alpha2 = dt / (den * dz)
#     kappa2 = torch.zeros_like(damp, device=device)
#     kappa2[:, 1:nx_pml - 2] = 0.5 * (damp[:, 1:nx_pml - 2] + damp[:, 2:nx_pml - 1]) * dt
    
#     kappa3 = torch.zeros_like(damp, device=device)
#     kappa3[free_surface_start:nz_pml - 2, :] = 0.5 * (damp[free_surface_start:nz_pml - 2, :] + damp[free_surface_start + 1:nz_pml - 1, :]) * dt
    
#     k = 0
#     for i, chunk in enumerate(torch.chunk(src_v, checkpoint_segments, dim=-1)):
#         # Step forward
#         p, u, w, rcv_p_temp, rcv_u_temp, rcv_w_temp, forward_wavefield_p_temp, forward_wavefield_u_temp, forward_wavefield_w_temp = \
#             checkpoint(step_forward,
#                        nx, nz, dx, dz, dt,
#                        nabc, free_surface,
#                        src_x, src_z, src_n, chunk,
#                        rcv_x, rcv_z, rcv_n,
#                        kappa1, alpha1, kappa2, alpha2, kappa3, c1_staggered, c2_staggered,
#                        p, u, w,
#                        device, dtype)

#         # Save the waveform recorded on the receiver
#         rcv_p[:, k:k + chunk.shape[-1]] = rcv_p_temp
#         rcv_u[:, k:k + chunk.shape[-1]] = rcv_u_temp
#         rcv_w[:, k:k + chunk.shape[-1]] = rcv_w_temp

#         # Accumulate the forward wavefield
#         forward_wavefield_p = forward_wavefield_p + forward_wavefield_p_temp.detach()
#         forward_wavefield_u = forward_wavefield_p + forward_wavefield_u_temp.detach()
#         forward_wavefield_w = forward_wavefield_p + forward_wavefield_w_temp.detach()
            
#         k = k + chunk.shape[-1]
    
#     record_waveform = {
#         "p": rcv_p,
#         "u": rcv_u,
#         "w": rcv_w,
#         "forward_wavefield_p": forward_wavefield_p,
#         "forward_wavefield_u": forward_wavefield_u,
#         "forward_wavefield_w": forward_wavefield_w,
#     }
    
#     return record_waveform


'''
* Author: LiuFeng(SJTU) : liufeng2317@sjtu.edu.cn
* Date: 2024-05-12 09:09:07
* LastEditors: LiuFeng
* LastEditTime: 2024-05-12 09:09:14
* Description: 
* Copyright (c) 2024 by liufeng, Email: liufeng2317@sjtu.edu.cn, All Rights Reserved.
'''
from .base import Misfit
import torch
import math
from .GlobalCorrelation import Misfit_global_correlation
from .Envelope import Misfit_envelope

class Misfit_weighted_ECI(Misfit):
    """Weighted envelope misfit and global correlation misfit (Song Chao et al., 2023, IEEE TGRS)
    
    Paraemters:
    -----------
        dt (float)      : time sampling interval
        p (float)       : the norm order of the reuslt
        instaneous_phase (bool) : use instaeous phase or amplitude for the misfit
        obs (Tensors)   : the observed waveform 
        syn (Tensors)   : the synthetic waveform 
    """
    def __init__(self,max_iter=1000,dt=1,p=1.5,instaneous_phase=False) -> None:
        super().__init__()
        self.max_iter = max_iter
        self.iter = 0
        self.dt = dt
        self.p  = p
        self.instaneous_phase   = instaneous_phase
        self.GC_fn              = Misfit_global_correlation(dt=self.dt)
        self.Envelope_fn        = Misfit_envelope(dt=self.dt,p=self.p,instaneous_phase=self.instaneous_phase)
        
    def forward(self,obs,syn):
        N = self.max_iter
        w_i = 1/(1+math.exp(-(self.iter - N/2)))
        GCN_loss = self.GC_fn.forward(obs=obs,syn=syn)
        ECI_loss = self.Envelope_fn.forward(obs=obs,syn=syn)
        loss = w_i*GCN_loss + (1-w_i)*ECI_loss
        self.iter += 1
        return loss


'''
* Author: LiuFeng(SJTU) : liufeng2317@sjtu.edu.cn
* Date: 2024-04-19 22:07:06
* LastEditors: LiuFeng
* LastEditTime: 2024-05-10 14:53:29
* Description: 
* Copyright (c) 2024 by liufeng, Email: liufeng2317@sjtu.edu.cn, All Rights Reserved.
'''

import numpy as np
import matplotlib.pyplot as plt
from ADFWI.utils.utils import gpu2cpu
import warnings
warnings.filterwarnings("ignore")

########################################################
#              waveform （single Trace）
########################################################
def plot_waveform_trace(data,shot=0,trace=0,dt=None,figsize=(12,4),color='k',
                    tick_param     = {'labelsize':18},label_param = {'family':'Times New Roman','weight':'normal','size': 18},
                    title_param = {'family':'Times New Roman','weight':'normal','size': 20},
                    show=True,save_path="",save_dpi=300):
    """plot single trace waveform
    Parameters
    ----------------
        - data (ndarray or Tensors)     : 3D waveform data
        - shot (int)                    : shot number 
        - trace (int)                   : trace number
        - color (str)                   : the color of lines
        - tick_param (dict,optional)    : the fontdict for ticks
        - label_param (dict,optional)   : the fontdict for label
        - title_param (dict,optional)   : the fontdict for title
        - figsize (tuple,optional)      : The size of the figure
        - show (bool,optional)          : showing the figure or not. Default True
        - save_path (str,optional)      : the saving path for the figure. Defualt:""
        - save_dpi (int,optional)       : the saving resolution for the figure. Default:300 dpi
    """
    data = gpu2cpu(data)
    
    plt.figure(figsize=figsize)
    if dt is not None:
        plt.plot(np.arange(data.shape[1])*dt,data[shot,:,trace],c=color)
        plt.xlabel("Times (s)",**label_param)
    else:
        plt.plot(data[shot,:,trace],c=color)
        plt.xlabel("Time Samples",**label_param)
    
    plt.tick_params(**tick_param)
    plt.title(f"shot:{shot} trace:{trace}",**title_param)
    
    plt.ylabel("Amplitude",**label_param)
    if not save_path == "":
        plt.savefig(save_path,dpi=save_dpi,bbox_inches="tight")
    if show:
        plt.show()
    else:
        plt.close()
    

########################################################
#                   waveform 2D
########################################################
def norm_traces(data):
    """ Normalize a seismic data to its maximum amplitude trace by trace
    
    Args:
        data (ndarray): The seismic data to be normalized.
        
    Returns:
        ndarray: The normalized seismic data.
    """

    eps = 1e-20

    nr = data.shape[0]
    nt = data.shape[1]


    for i in range(nr):
        data[i,:] = data[i,:]/(abs(data[i,:]) + eps).max()

    return data

def plot_waveform2D(data, dt=None, dx=None, cmap='coolwarm', aspect='auto', clip=99.9, 
                figsize=(10,6), colorbar=False,
                type='section', norm=False, wiggle_scale=1, wiggle_interval=1,
                plot_fk=False, vel=None, fmin=1, fmax=20, kmin=-0.05, kmax=0.05,
                save_path="",
                show=False
                ):
    """ Plot two data
    
    Plot a data section using matplotlib.imshow.
    
    Parameters:
        - data (ndarray or Tensor): The data to be plotted.
        - dt (float, optional): The time sampling interval. If not provided, the x-axis will be labeled with sample numbers.
        - dx (float, optional): The spatial sampling interval. If not provided, the y-axis will be labeled with trace numbers.
        - cmap (str, optional): The colormap to be used. Default is 'gray'.
        - aspect (str, optional): The aspect ratio of the plot. Default is 'auto'.
        - clip (float, optional): The percentile value for clipping the data. Default is 99.9.
        - figsize (tuple, optional): The size of the figure. Default is (8, 8).
        - colorbar (bool, optional): Whether to show the colorbar. Default is False.
        - norm (bool, optional): Whether to normalize the data. Default is False.
        - savefig (str, optional): The file path to save the figure. If not provided, the figure will be displayed.

    Returns:
    None
    """

    plt.figure(figsize = figsize)
    
    if not isinstance(data, np.ndarray):
        try:
            data = data.cpu().detach().numpy()
        except AttributeError:
            # Handle the case where data cannot be converted to a NumPy array
            pass    
    if dt is None:
        t = np.arange(data.shape[1])
    else:
        t = np.arange(data.shape[1]) * dt
    if dx is None:
        x = np.arange(data.shape[0])
    else:
        x = np.arange(data.shape[0]) * dx

    extent = [x[0], x[-1], t[-1], t[0]]

    if plot_fk:
        plt.subplot(1,2,1)
    else:
        plt.subplot(1,1,1)

    if norm:
        data = norm_traces(data)

    vmax = np.percentile(data, clip)
    if type == 'section':
        plt.imshow(data.T, aspect=aspect, cmap=cmap, vmin=-vmax, vmax=vmax, extent = extent)
        plt.ylim([t[-1], t[0]])
        plt.xlim([x[0], x[-1]])
    
    elif type == 'wiggle':
        for i, trace in enumerate(data):
            if i % wiggle_interval != 0:
                continue
            trace = trace * wiggle_scale + i
            plt.plot(trace, t, color='black', linewidth=1.0)
            plt.fill_betweenx(t, i, trace, where=(trace > i), color='black')
        plt.ylim([t[-1], t[0]])
        plt.xlim([0-1, data.shape[0]+1])
    else:
        raise ValueError('type must be either "section" or "wiggle"')

    # lim
    if dx is not None:
        plt.xlabel('Offset (m)')
    else:
        plt.xlabel('Trace #')

    if dt is not None:
        plt.ylabel('Time (s)')
    else:
        plt.ylabel('Sample #')
    if colorbar:
        plt.colorbar()
    plt.grid(axis='y', alpha=0.8)
    plt.tight_layout()
    
    if not save_path == "":
        plt.savefig(save_path, dpi=300,bbox_inches="tight")
    
    if not show:
        plt.close()
    else:
        plt.show()

########################################################
#                waveform wiggle 2D
########################################################
def insert_zeros(trace, tt=None):
    """Insert zero locations in data trace and tt vector based on linear fit"""

    if tt is None:
        tt = np.arange(len(trace))

    # Find zeros
    zc_idx = np.where(np.diff(np.signbit(trace)))[0]
    x1 = tt[zc_idx]
    x2 = tt[zc_idx + 1]
    y1 = trace[zc_idx]
    y2 = trace[zc_idx + 1]
    a = (y2 - y1) / (x2 - x1)
    tt_zero = x1 - y1 / a

    # split tt and trace
    tt_split = np.split(tt, zc_idx + 1)
    trace_split = np.split(trace, zc_idx + 1)
    tt_zi = tt_split[0]
    trace_zi = trace_split[0]

    # insert zeros in tt and trace
    for i in range(len(tt_zero)):
        tt_zi = np.hstack(
            (tt_zi, np.array([tt_zero[i]]), tt_split[i + 1]))
        trace_zi = np.hstack(
            (trace_zi, np.zeros(1), trace_split[i + 1]))

    return trace_zi, tt_zi


def wiggle_input_check(data, tt, xx, sf, verbose):
    ''' Helper function for wiggle() and traces() to check input

    '''
    # Input check for verbose
    if not isinstance(verbose, bool):
        raise TypeError("verbose must be a bool")

    # Input check for data
    if type(data).__module__ != np.__name__:
        raise TypeError("data must be a numpy array")

    if len(data.shape) != 2:
        raise ValueError("data must be a 2D array")

    # Input check for tt
    if tt is None:
        tt = np.arange(data.shape[0])
        if verbose:
            print("tt is automatically generated.")
            print(tt)
    else:
        if type(tt).__module__ != np.__name__:
            raise TypeError("tt must be a numpy array")
        if len(tt.shape) != 1:
            raise ValueError("tt must be a 1D array")
        if tt.shape[0] != data.shape[0]:
            raise ValueError("tt must have same as data's rows")

    # Input check for xx
    if xx is None:
        xx = np.arange(data.shape[1])
        if verbose:
            print("xx is automatically generated.")
            print(xx)
    else:
        if type(xx).__module__ != np.__name__:
            raise TypeError("tt must be a numpy array")
        if len(xx.shape) != 1:
            raise ValueError("tt must be a 1D array")
        if tt.shape[0] != data.shape[0]:
            raise ValueError("tt must have same as data's rows")
        if verbose:
            print(xx)

    # Input check for streth factor (sf)
    if not isinstance(sf, (int, float)):
        raise TypeError("Strech factor(sf) must be a number")

    # Compute trace horizontal spacing
    ts = np.min(np.diff(xx))

    # Rescale data by trace_spacing and strech_factor
    data_max_std = np.max(np.std(data, axis=0))
    data = data / data_max_std * ts * sf

    return data, tt, xx, ts


def plot_waveform_wiggle(data, tt=None, xx=None, color='k', sf=0.15, verbose=False,save_path="",show=False):
    '''Wiggle plot of a sesimic data section
    Syntax examples:
        wiggle(data)
        wiggle(data, tt)
        wiggle(data, tt, xx)
        wiggle(data, tt, xx, color)
        fi = wiggle(data, tt, xx, color, sf, verbose)
    Parameters
    ----------------------
        - data (ndarray or Tensors)         : 2D waveform data
        - tt (ndarray or Tensors,optional)  : time list of the waveforms
        - xx (ndarray or Tensors,optional)  : offset list of the waveforms
        - color (str,optional)              : color of the waveform
        - sf (floats,optional)              : streth factor
        - verbose (bool,optional)           : show the ouputs or not
        - save_path (str,optional)          : save path for the figure
        - show (bool,optional)              : show the figure or not

    The following color abbreviations are supported:

    ==========  ========
    character   color
    ==========  ========
    'b'         blue
    'g'         green
    'r'         red
    'c'         cyan
    'm'         magenta
    'y'         yellow
    'k'         black
    'w'         white
    ==========  ========

    returns:
    --------------------
    None
    '''
    data = gpu2cpu(data)
    if tt is not None:
        tt = gpu2cpu(tt)
    if xx is not None:
        xx = gpu2cpu(xx)
    # Input check
    data, tt_new, xx_new, ts_new = wiggle_input_check(data, tt, xx, sf, verbose)

    # Plot data using matplotlib.pyplot
    Ntr = data.shape[1]

    ax = plt.gca()
    for ntr in range(Ntr):
        trace = data[:, ntr]
        offset = xx_new[ntr]

        if verbose:
            print(offset)

        trace_zi, tt_zi = insert_zeros(trace, tt_new)
        ax.fill_betweenx(tt_zi, offset, trace_zi + offset,
                        where=trace_zi >= 0,
                        facecolor=color)
        ax.plot(trace_zi + offset, tt_zi, color)

    ax.set_xlim(xx_new[0] - ts_new, xx_new[-1] + ts_new)
    ax.set_ylim(tt_new[0], tt_new[-1])
    ax.invert_yaxis()
    
    if xx is not None:
        plt.xlabel('Offset (m)')
    else:
        plt.xlabel('Trace #')

    if tt is not None:
        plt.ylabel('Time (s)')
    else:
        plt.ylabel('Sample #')
    
    if save_path != "":
        plt.savefig(save_path,bbox_inches="tight")
    if show:
        plt.show()
    else:
        plt.close()


from .dip_acoustic_model import DIP_AcousticModel
from .dip_acoustic_fwi import DIP_AcousticFWI
from .dip_elastic_model import DIP_ElasticModel
from .dip_vti_model import DIP_VTIModel
from .dip_elastic_fwi import DIP_ElasticFWI

from .model.Unet import UNet as DIP_Unet
from .model.CNN import  CNN as DIP_CNN
from .model.ResNet import ResNet as DIP_ResNet
from .model.MLP import MLP as DIP_MLP
from .model.MobileNetV2 import MobileNetV2FWI as DIP_MobilNetV2


'''
* Author: LiuFeng(SJTU) : liufeng2317@sjtu.edu.cn
* Date: 2024-04-26 19:42:24
* LastEditors: LiuFeng
* LastEditTime: 2024-05-13 23:16:49
* Description: 
* Copyright (c) 2024 by liufeng, Email: liufeng2317@sjtu.edu.cn, All Rights Reserved.
'''
from typing import Optional,Union,List
import os
import math
import torch
import numpy as np
from tqdm import tqdm
from ADFWI.model       import AbstractModel
from ADFWI.dip.dip_vti_model import DIP_VTIModel
from ADFWI.dip.dip_vti_model_o4a import DIP_VTIModel_o4a
from ADFWI.propagator  import ElasticPropagator,GradProcessor
from ADFWI.survey      import SeismicData
from ADFWI.fwi.misfit  import Misfit,Misfit_NIM
from ADFWI.fwi.regularization import Regularization
from ADFWI.utils       import numpy2tensor
from ADFWI.view        import plot_vp_vs_rho,plot_model,plot_eps_delta_gamma
from ADFWI.fwi.multiScaleProcessing import lpass

class DIP_ElasticFWI(torch.nn.Module):
    """Elastic Full waveform inversion class
    """
    def __init__(self,propagator:ElasticPropagator,model:AbstractModel,
                 loss_fn:Union[Misfit,torch.autograd.Function],
                 obs_data:SeismicData,
                 optimizer:Union[torch.optim.Optimizer,List[torch.optim.Optimizer]]      = None,
                 scheduler:torch.optim.lr_scheduler                                      = None,
                 gradient_processor: Union[GradProcessor,List[GradProcessor]]            = None,
                 regularization_fn:Optional[Regularization]                              = None,
                 regularization_weights_x:Optional[List[Union[float]]]                   = [0,0,0,0,0,0],       # vp/vs/rho epsilon/delta/gamma
                 regularization_weights_z:Optional[List[Union[float]]]                   = [0,0,0,0,0,0],       # vp/vs/rho epsilon/delta/gamma
                 waveform_normalize:Optional[bool]                                       = True,
                 cache_result:Optional[bool]                                             = True,
                 save_fig_epoch:Optional[int]                                            = -1,
                 save_fig_path:Optional[str]                                             = "",
                 inversion_component:Optional[np.array]                                   = ["pressure"],
                ):
        """
        Parameters:
        --------------
        propagator (ElasticPropagator)                                          : The propagator used for simulating elastic wave propagation.
        model (AbstractModel)                                                   : The model class representing the velocity structure.
        loss_fn (Union[Misfit, torch.autograd.Function])                        : The loss function used to compute the misfit between observed and predicted data.
        obs_data (SeismicData)                                                  : The observed seismic data.
        optimizer (Union[torch.optim.Optimizer, List[torch.optim.Optimizer]])   : The optimizer or list of optimizers for model parameters. Default is None.
        scheduler (Optional[torch.optim.lr_scheduler])                          : The learning rate scheduler for optimizing the model parameters. Default is None.
        gradient_processor (Union[GradProcessor, List[GradProcessor]])          : Processor(s) for handling gradients (e.g., vp/vs/rho, epsilon/delta/gamma). Default is None.
        regularization_fn (Optional[Regularization])                            : Regularization function(s) applied to parameters like vp/vs/rho/epsilon/delta/gamma. Default is None.
        regularization_weights_x (Optional[List[Union[float]]])                 : Regularization weights for the x-axis. Default is [0, 0, 0, 0, 0, 0].
        regularization_weights_z (Optional[List[Union[float]]])                 : Regularization weights for the z-axis. Default is [0, 0, 0, 0, 0, 0].
        waveform_normalize (Optional[bool])                                     : Whether to normalize the waveforms during inversion. Default is True.
        cache_result (Optional[bool])                                           : Whether to save intermediate results during the inversion. Default is True.
        cache_gradient (Optional[bool])                                         : Whether to save model variations (not gradients) during inversion. Default is False.
        save_fig_epoch (Optional[int])                                          : The interval (in epochs) at which to save the inversion result figure. Default is -1 (no figure saved).
        save_fig_path (Optional[str])                                           : The path where to save the inversion result figure. Default is an empty string (no save path).
        inversion_component (Optional[np.array])                                : The components of the inversion (e.g., ["pressure"]). Default is ["pressure"].
        """
        super().__init__()
        self.propagator                 = propagator
        self.model                      = model
        self.optimizer                  = optimizer
        self.scheduler                  = scheduler
        self.loss_fn                    = loss_fn
        self.regularization_fn          = regularization_fn
        self.regularization_weights_x   = regularization_weights_x
        self.regularization_weights_z   = regularization_weights_z
        self.obs_data                   = obs_data
        self.gradient_processor         = gradient_processor
        self.device                     = self.propagator.device
        self.dtype                      = self.propagator.dtype 
        
        # receiver masks
        receiver_masks = self.propagator.receiver_masks
        if receiver_masks is None:
            receiver_masks = np.ones((self.propagator.src_n,self.propagator.rcv_n))
        receiver_masks = numpy2tensor(receiver_masks).to(self.device)
        self.receiver_masks = receiver_masks.unsqueeze(1).expand(-1, self.propagator.nt, -1)  # [shot, time, rcv]

        # observed data
        self.waveform_normalize = waveform_normalize
        obs_p   = -(self.obs_data.data["txx"]+self.obs_data.data["tzz"])
        obs_p   = (numpy2tensor(obs_p,self.dtype).to(self.device))*self.receiver_masks
        obs_vx  = (numpy2tensor(self.obs_data.data["vx"],self.dtype).to(self.device))*self.receiver_masks
        obs_vz  = (numpy2tensor(self.obs_data.data["vz"],self.dtype).to(self.device))*self.receiver_masks
        if self.waveform_normalize:
            obs_p  = self._normalize(obs_p)
            obs_vx = self._normalize(obs_vx)
            obs_vz = self._normalize(obs_vz)
        self.obs_p = obs_p
        self.obs_vx = obs_vx
        self.obs_vz = obs_vz
        
        # save result
        self.cache_result   = cache_result
        self.iter_vp,self.iter_vs,self.iter_rho = [],[],[]       
        self.iter_eps,self.iter_delta,self.iter_gamma = [],[],[]
        self.iter_loss      = []
        
        # optimizer
        if not isinstance(self.optimizer, list):
            self.optimizer = [self.optimizer]
        
        if not isinstance(self.scheduler, list):
            self.scheduler = [self.scheduler]
            
        
        # save figure
        self.save_fig_epoch = save_fig_epoch
        self.save_fig_path  = save_fig_path
        
        # inversion component
        self.inversion_component = inversion_component

    def _normalize(self,data):
        mask    = torch.sum(torch.abs(data),axis=1,keepdim=True) == 0
        max_val = torch.max(torch.abs(data),axis=1,keepdim=True).values
        max_val = max_val.masked_fill(mask, 1)
        data = data/max_val
        return data
    
    # misfits calculation
    def calculate_loss(self, synthetic_waveform, observed_waveform, normalization, loss_fn, cutoff_freq=None, propagator_dt=None):
        """
        Generalized function to calculate misfit loss for a given component.
        """
        if normalization:
            synthetic_waveform = synthetic_waveform / (torch.max(torch.abs(synthetic_waveform), axis=1, keepdim=True).values)
        # Apply low-pass filter if cutoff frequency is provided
        if cutoff_freq is not None:
            synthetic_waveform, observed_waveform = lpass(synthetic_waveform, observed_waveform, cutoff_freq, int(1 / propagator_dt))
        if isinstance(loss_fn, Misfit):
            return loss_fn.forward(synthetic_waveform, observed_waveform)
        else:
            return loss_fn.apply(synthetic_waveform, observed_waveform)
    
    # regularization calculation
    def calculate_regularization_loss(self, model_param, weight_x, weight_z, regularization_fn):
        """
        Generalized function to calculate regularization loss for a given parameter.
        """
        regularization_loss = torch.tensor(0.0, device=model_param.device)
        # Check if the parameter requires gradient
        if model_param.requires_grad:
            # Set the regularization weights for x and z directions
            regularization_fn.alphax = weight_x
            regularization_fn.alphaz = weight_z
            # Calculate regularization loss if any weight is greater than zero
            if regularization_fn.alphax > 0 or regularization_fn.alphaz > 0:
                regularization_loss = regularization_fn.forward(model_param)
        return regularization_loss
    
        # gradient precondition
    def process_gradient(self, parameter,forw,idx=None):
        with torch.no_grad():
            grads = parameter.grad.cpu().detach().numpy()
            vmax = np.max(parameter.cpu().detach().numpy())
            # Apply gradient processor
            if isinstance(self.gradient_processor, GradProcessor):
                grads = self.gradient_processor.forward(nz=self.model.nz, nx=self.model.nx, vmax=vmax, grad=grads, forw=forw)
            else:
                grads = self.gradient_processor[idx].forward(nz=self.model.nz, nx=self.model.nx, vmax=vmax, grad=grads, forw=forw)
            # Convert grads back to tensor and assign
            grads_tensor = numpy2tensor(grads, dtype=self.propagator.dtype).to(self.propagator.device)
            parameter.grad = grads_tensor
    
    def save_vp_vs_rho_fig(self,epoch_id,vp,vs,rho):
        vp_bound    =  self.model.get_bound("vp")
        vs_bound    =  self.model.get_bound("vs")
        rho_bound   =  self.model.get_bound("rho")
        if vp_bound[0] is None and vp_bound[1] is None:
            self.vp_min = self.model.get_model("vp").min() - 500
            self.vp_max = self.model.get_model("vp").max() + 500
        else: 
            self.vp_min = vp_bound[0]
            self.vp_max = vp_bound[1]
            if self.model.water_layer_mask is not None:
                self.vp_min = 1500
        if vs_bound[0] is None and vs_bound[1] is None:
            self.vs_min = self.model.get_model("vs").min() - 500
            self.vs_max = self.model.get_model("vs").max() + 500
        else: 
            self.vs_min = vs_bound[0]
            self.vs_max = vs_bound[1]
            if self.model.water_layer_mask is not None:
                self.vs_min = 0        
        if rho_bound[0] is None and rho_bound[1] is None:
            self.rho_min = self.model.get_model("rho").min() - 200
            self.rho_max = self.model.get_model("rho").max() + 200
        else: 
            self.rho_min = rho_bound[0]
            self.rho_max = rho_bound[1]
            if self.model.water_layer_mask is not None:
                self.rho_min = 1000
        
        if self.save_fig_epoch == -1:
            pass
        elif epoch_id%self.save_fig_epoch == 0:
            if os.path.exists(self.save_fig_path):
                plot_vp_vs_rho(
                    vp=vp,vs=vs,rho=rho,
                    # title=f"Iteration {i}",
                    figsize=(12,5),wspace=0.2,cbar_pad_fraction=0.18,cbar_height=0.04,
                    dx=self.model.dx,dz=self.model.dz,
                    vp_min=self.vp_min,vp_max=self.vp_max,
                    vs_min=self.vs_min,vs_max=self.vs_max,
                    rho_min=self.rho_min,rho_max=self.rho_max,
                    save_path=os.path.join(self.save_fig_path,f"model_{epoch_id}.png"),
                    show=False
                    )
        return

    def save_eps_delta_gamma_fig(self,epoch_id,eps,delta,gamma):
            eps_bound    =  self.model.get_bound("eps")
            delta_bound    =  self.model.get_bound("delta")
            gamma_bound   =  self.model.get_bound("gamma")
            if eps_bound[0] is None and eps_bound[1] is None:
                self.vp_min = self.model.get_model("eps").min() - 0.01
                self.vp_max = self.model.get_model("eps").max() + 0.01
            else: 
                self.vp_min = eps_bound[0]
                self.vp_max = eps_bound[1]
            
            if delta_bound[0] is None and delta_bound[1] is None:
                self.delta_min = self.model.get_model("delta").min() - 0.01
                self.delta_max = self.model.get_model("delta").max() + 0.01
            else: 
                self.vs_min = delta_bound[0]
                self.vs_max = delta_bound[1]
            
            if gamma_bound[0] is None and gamma_bound[1] is None:
                self.gamma_min = self.model.get_model("gamma").min() - 0.01
                self.gamma_max = self.model.get_model("gamma").max() + 0.01
            else: 
                self.rho_min = gamma_bound[0]
                self.rho_max = gamma_bound[1]
        
            if self.save_fig_epoch == -1:
                pass
            elif epoch_id%self.save_fig_epoch == 0:
                if os.path.exists(self.save_fig_path):
                    plot_eps_delta_gamma(
                        eps=eps,delta=delta,gamma=gamma,
                        # title=f"Iteration {i}",
                        figsize=(12,5),wspace=0.3,cbar_pad_fraction=0.01,cbar_height=0.04,
                        dx=self.model.dx,dz=self.model.dz,
                        save_path=os.path.join(self.save_fig_path,f"anisotropic_model_{epoch_id}.png"),
                        show=False
                        )
            return
    
    def save_model(self,epoch_id,loss_epoch):
        """
            Save model parameters and gradients if caching is enabled.
        """
        # Save the loss
        self.iter_loss.append(loss_epoch)

        # Save the model parameters
        param_names = ["vp", "vs", "rho"]
        anisotropic_params = ["eps", "delta", "gamma"] if isinstance(self.model, (DIP_VTIModel,DIP_VTIModel_o4a)) else []
        for name in param_names + anisotropic_params:
            param = getattr(self.model, name, None)
            if param is not None:
                temp_param = param.cpu().detach().numpy()
                getattr(self, f"iter_{name}").append(temp_param)
        
        # save the figure
        self.save_vp_vs_rho_fig(epoch_id,self.model.vp.cpu().detach().numpy(),
                                         self.model.vs.cpu().detach().numpy(),
                                         self.model.rho.cpu().detach().numpy())
        if isinstance(self.model,(DIP_VTIModel,DIP_VTIModel_o4a)):
            self.save_eps_delta_gamma_fig(epoch_id,
                                          self.model.eps.cpu().detach().numpy(),
                                          self.model.delta.cpu().detach().numpy(),
                                          self.model.gamma.cpu().detach().numpy())
        return

    def forward(self,
                iteration:int,
                fd_order:int                        = 4,
                batch_size:Optional[int]            = None,
                checkpoint_segments:Optional[int]   = 1 ,
                start_iter                          = 0,
                cutoff_freq                         = None,
                ):
        """
        Parameters:
        ------------
        iteration (int)                     : The maximum iteration number in the inversion process.
        fd_order (int)                      : The order of the finite difference scheme for wave propagation. Default is 4.
        batch_size (Optional[int])          : The number of shots (data samples) in each batch. Default is None, meaning use all available shots.
        checkpoint_segments (Optional[int]) : The number of segments into which the time series should be divided for memory efficiency. Default is 1, which means no segmentation.
        start_iter (int)                    : The starting iteration for the optimization process (e.g., for optimizers like Adam/AdamW, and learning rate schedulers like step_lr). Default is 0.
        cutoff_freq (Optional[float])       : The cutoff frequency for low-pass filtering, if specified. Default is None (no filtering applied).
        """
        n_shots = self.propagator.src_n
        if batch_size is None or batch_size > n_shots:
            batch_size = n_shots
        
        # epoch
        pbar_epoch = tqdm(range(start_iter,start_iter+iteration),position=0,leave=False,colour='green',ncols=80)
        for i in pbar_epoch:
            # batch
            for opt in self.optimizer:
                opt.zero_grad()
            loss_epoch = 0
            pbar_batch = tqdm(range(math.ceil(n_shots/batch_size)),position=1,leave=False,colour='red',ncols=80)
            for batch in pbar_batch:
                # forward simulation
                begin_index = 0  if batch==0 else batch*batch_size
                end_index   = n_shots if batch==math.ceil(n_shots/batch_size)-1 else (batch+1)*batch_size
                shot_index  = np.arange(begin_index,end_index)
                record_waveform = self.propagator.forward(fd_order=fd_order,shot_index=shot_index,checkpoint_segments=checkpoint_segments)
                rcv_txx,rcv_tzz,rcv_txz,rcv_vx,rcv_vz = record_waveform["txx"],record_waveform["tzz"],record_waveform["txz"],record_waveform["vx"],record_waveform["vz"]
                forward_wavefield_txx,forward_wavefield_tzz,forward_wavefield_txz,forward_wavefield_vx,forward_wavefield_vz = record_waveform["forward_wavefield_txx"],record_waveform["forward_wavefield_tzz"],record_waveform["forward_wavefield_txz"],record_waveform["forward_wavefield_vx"],record_waveform["forward_wavefield_vz"]
                
                # misfits
                loss_pressure, loss_vx, loss_vz = 0, 0, 0
                receiver_mask = self.receiver_masks[shot_index]
                if "pressure" in self.inversion_component:
                    if batch == 0:
                        forw  = -(forward_wavefield_txx + forward_wavefield_tzz).cpu().detach().numpy()
                    else:
                        forw += -(forward_wavefield_txx + forward_wavefield_tzz).cpu().detach().numpy()
                    syn_p = -(rcv_txx + rcv_tzz)
                    syn_p = syn_p*receiver_mask
                    loss_pressure = self.calculate_loss(syn_p, self.obs_p[shot_index], self.waveform_normalize, self.loss_fn, cutoff_freq, self.propagator.dt)
                if "vx" in self.inversion_component:
                    forw = forward_wavefield_vx.cpu().detach().numpy()
                    rcv_vx = rcv_vx*receiver_mask
                    loss_vx = self.calculate_loss(rcv_vx, self.obs_vx[shot_index],self.waveform_normalize, self.loss_fn, cutoff_freq, self.propagator.dt)
                if "vz" in self.inversion_component:
                    rcv_vz = rcv_vz*receiver_mask
                    forw = forward_wavefield_vz.cpu().detach().numpy()
                    loss_vz = self.calculate_loss(rcv_vz, self.obs_vz[shot_index],self.waveform_normalize, self.loss_fn, cutoff_freq, self.propagator.dt)
                data_loss = loss_pressure + loss_vx + loss_vz
                
                # regularization
                if self.regularization_fn is not None:
                    # Initialize regularization losses
                    regularization_loss_vp  = self.calculate_regularization_loss(self.model.vp , self.regularization_weights_x[0], self.regularization_weights_z[0], self.regularization_fn)
                    regularization_loss_vs  = self.calculate_regularization_loss(self.model.vs , self.regularization_weights_x[1], self.regularization_weights_z[1], self.regularization_fn)
                    regularization_loss_rho = self.calculate_regularization_loss(self.model.rho, self.regularization_weights_x[2], self.regularization_weights_z[2], self.regularization_fn)
                    # For anisotropic model parameters
                    regularization_loss_eps = regularization_loss_delta = regularization_loss_gamma = torch.tensor(0.0, device=self.device)
                    if isinstance(self.model, (DIP_VTIModel,DIP_VTIModel_o4a)):
                        regularization_loss_eps   = self.calculate_regularization_loss(self.model.eps  , self.regularization_weights_x[3], self.regularization_weights_z[3], self.regularization_fn)
                        regularization_loss_delta = self.calculate_regularization_loss(self.model.delta, self.regularization_weights_x[4], self.regularization_weights_z[4], self.regularization_fn)
                        regularization_loss_gamma = self.calculate_regularization_loss(self.model.gamma, self.regularization_weights_x[5], self.regularization_weights_z[5], self.regularization_fn)
                    # Summing all regularization losses
                    regularization_loss = (regularization_loss_vp + regularization_loss_vs + regularization_loss_rho +
                                        regularization_loss_eps + regularization_loss_delta + regularization_loss_gamma)
                    # Adding regularization loss to total loss
                    loss_epoch += data_loss.item() + regularization_loss.item()
                    loss = data_loss + regularization_loss
                else:
                    loss_epoch += data_loss.item()
                    loss = data_loss
                loss.backward()
                if math.ceil(n_shots/batch_size) == 1:
                    pbar_batch.set_description(f"Shot:{begin_index} to {end_index}")
            
            # gradient postprocess
            def grad_post_process(grads, parameter, forw=None, idx=None):
                grads = grads.cpu().detach().numpy()
                with torch.no_grad():
                    param = getattr(self.model, parameter).cpu().detach().numpy()
                    vmax = np.max(param)
                    # Apply gradient processor
                    if isinstance(self.gradient_processor, GradProcessor):
                        grads = self.gradient_processor.forward(nz=self.propagator.model.nz, nx=self.propagator.model.nx, vmax=vmax, grad=grads, forw=forw)
                    else:
                        grads = self.gradient_processor[idx].forward(nz=self.propagator.model.nz, nx=self.propagator.model.nx, vmax=vmax, grad=grads, forw=forw)
                    grads = numpy2tensor(grads, dtype=self.propagator.dtype).to(self.propagator.device)
                return grads

            # Register hooks for each model parameter
            if self.propagator.model.get_requires_grad("vp"):
                self.propagator.model.vp.register_hook(lambda grad: grad_post_process(grad, "vp", forw=forw, idx=0))
            if self.propagator.model.get_requires_grad("vs"):
                self.propagator.model.vs.register_hook(lambda grad: grad_post_process(grad, "vs", forw=forw, idx=1))
            if self.propagator.model.get_requires_grad("rho"):
                self.propagator.model.rho.register_hook(lambda grad: grad_post_process(grad, "rho", forw=forw, idx=2))
            if isinstance(self.model, (DIP_VTIModel,DIP_VTIModel_o4a)):
                if self.propagator.model.get_requires_grad("eps"):
                    self.propagator.model.eps.register_hook(lambda grad: grad_post_process(grad, "eps", forw=forw, idx=3))
                if self.propagator.model.get_requires_grad("delta"):
                    self.propagator.model.delta.register_hook(lambda grad: grad_post_process(grad, "delta", forw=forw, idx=4))
                if self.propagator.model.get_requires_grad("gamma"):
                    self.propagator.model.gamma.register_hook(lambda grad: grad_post_process(grad, "gamma", forw=forw, idx=5))
            
            for opt in self.optimizer:
                opt.step()
            
            for schdul in self.scheduler:
                schdul.step()
            
            # constrain the model parameters
            with torch.no_grad():
                self.propagator.model.forward()
            
            # cache results
            if self.cache_result:
                self.save_model(epoch_id=i,loss_epoch=loss_epoch)   
                
            pbar_epoch.set_description("Iter:{},Loss:{:.4}".format(i+1,loss_epoch))


'''
* Author: LiuFeng(SJTU) : liufeng2317@sjtu.edu.cn
* Date: 2024-04-20 09:32:43
* LastEditors: LiuFeng
* LastEditTime: 2024-05-15 19:30:38
* Description: 
* Copyright (c) 2024 by liufeng, Email: liufeng2317@sjtu.edu.cn, All Rights Reserved.
'''
import numpy as np
import torch
from torch import Tensor
from typing import Optional,Tuple,Union,List
from ADFWI.utils import gpu2cpu,numpy2tensor,tensor2numpy
from ADFWI.model.base import AbstractModel
from ADFWI.model.parameters import (thomsen_init,elastic_moduli_init,
                         vs_vp_to_Lame,thomsen_to_elastic_moduli,
                         elastic_moduli_for_isotropic,elastic_moduli_for_TI,
                         parameter_staggered_grid)
from ADFWI.view import (plot_vp_rho,plot_vp_vs_rho,plot_eps_delta_gamma,plot_lam_mu,plot_model)
from ADFWI.survey import Survey

class IsotropicElasticModel(AbstractModel):
    """Isotropic Elastic Velocity model with parameterization of vp vs and rho
    """
    def __init__(self,
                ox:float,oz:float,
                nx:int,nz:int,
                dx:float,dz:float,
                vp:Optional[Union[np.array,Tensor]]              = None,     # model parameter
                vs:Optional[Union[np.array,Tensor]]              = None,
                rho:Optional[Union[np.array,Tensor]]             = None,
                vp_bound: Optional[Tuple[float, float]]          = None,     # model parameter's boundary
                vs_bound: Optional[Tuple[float, float]]          = None,
                rho_bound: Optional[Tuple[float, float]]         = None,
                vp_grad:Optional[bool]                           = False,    # requires gradient or not
                vs_grad:Optional[bool]                           = False,
                rho_grad:Optional[bool]                          = False,
                free_surface:Optional[bool]                      = False,
                abc_type:Optional[str]                           = 'PML',
                abc_jerjan_alpha:Optional[float]                 = 0.0053,
                nabc:Optional[int]                               = 20,
                auto_update_rho:Optional[bool]                   = True,
                auto_update_vp:Optional[bool]                    = False,
                water_layer_mask:Optional[Union[np.array,Tensor]]= None,
                device                                           = 'cpu',
                dtype                                            = torch.float32
                )->None:
        """
        Parameters:
        --------------
        ox (float), oz (float)                   : Non use, the origin coordinates of the model in the x- and z- directions (in meters).
        nx (int), nz (int)                       : The number of grid points in the x- and z- directions.
        dx (float), dz (float)                   : The grid spacing in the x- and z- directions (in meters).
        vp (Optional[Union[np.array, Tensor]])   : P-wave velocity model with shape (nz, nx). Default is None.
        vs (Optional[Union[np.array, Tensor]])   : S-wave velocity model with shape (nz, nx). Default is None.
        rho (Optional[Union[np.array, Tensor]])  : Density model with shape (nz, nx). Default is None.
        vp_bound (Optional[Tuple[float, float]]) : The lower and upper bounds for the P-wave velocity model. Default is None.
        vs_bound (Optional[Tuple[float, float]]) : The lower and upper bounds for the S-wave velocity model. Default is None.
        rho_bound (Optional[Tuple[float, float]]): The lower and upper bounds for the density model. Default is None.
        vp_grad (Optional[bool])                 : A flag to indicate if the gradient of the P-wave velocity model is needed. Default is False.
        vs_grad (Optional[bool])                 : A flag to indicate if the gradient of the S-wave velocity model is needed. Default is False.
        rho_grad (Optional[bool])                : A flag to indicate if the gradient of the density model is needed. Default is False.
        free_surface (Optional[bool])            : A flag to indicate the presence of a free surface in the model. Default is False.
        abc_type (Optional[str])                 : The type of absorbing boundary condition used in the model. Options include 'PML', 'Jerjan', etc. Default is 'PML'.
        abc_jerjan_alpha (Optional[float])       : The attenuation factor for the Jerjan boundary condition. Default is 0.0053.
        nabc (Optional[int])                     : The number of grid cells dedicated to the absorbing boundary. Default is 20.
        auto_update_rho (Optional[bool])         : Whether to automatically update the density model during inversion. Default is True.
        auto_update_vp (Optional[bool])          : Whether to automatically update the P-wave velocity model during inversion. Default is False.
        water_layer_mask (Optional[Union[np.array, Tensor]]) : A mask for the water layer (not update), if applicable. Default is None.
        device (str)                             : The device on which to run the model. Options are 'cpu' or 'cuda'. Default is 'cpu'.
        dtype (torch.dtype)                      : The data type for PyTorch tensors. Default is torch.float32.
        """
        # initialize the common model parameters
        super().__init__(ox,oz,nx,nz,dx,dz,free_surface,abc_type,abc_jerjan_alpha,nabc,device,dtype)

        # initialize the thomson model model parameters
        self.pars       = ["vp","vs","rho"]
        self.vp         = vp.copy()
        self.vs         = vs.copy()
        self.rho        = rho.copy()
        self.vp_grad    = vp_grad
        self.vs_grad    = vs_grad
        self.rho_grad   = rho_grad
        self.eps        = np.zeros((nz,nx))
        self.gamma      = np.zeros((nz,nx))
        self.delta      = np.zeros((nz,nx))
        self._parameterization_thomson()
        
        # initialize the lame constant
        self.mu         = None
        self.lamu       = None
        self.lam        = None
        self.muxz       = None
        self.b          = None
        self.bx         = None
        self.bz         = None
        self._parameterization_Lame()
        
        # initialize the elastic moduli
        self.CC         = []
        self._parameterization_elastic_moduli()
        
        # set model bounds
        self.lower_bound["vp"]  = vp_bound[0]  if vp_bound  is not None else None
        self.lower_bound["vs"]  = vs_bound[0]  if vs_bound  is not None else None
        self.lower_bound["rho"] = rho_bound[0] if rho_bound is not None else None
        self.upper_bound["vp"]  = vp_bound[1]  if vp_bound  is not None else None
        self.upper_bound["vs"]  = vs_bound[1]  if vs_bound  is not None else None
        self.upper_bound["rho"] = rho_bound[1] if rho_bound is not None else None
        
        # set model gradients
        self.requires_grad["vp"]    = self.vp_grad
        self.requires_grad["vs"]    = self.vs_grad
        self.requires_grad["rho"]   = self.rho_grad
        
        # check the input model
        self._check_bounds()
        self.check_dims()
        
        # update rho using the empirical function
        self.auto_update_rho = auto_update_rho
        self.auto_update_vp  = auto_update_vp
    
        if water_layer_mask is not None:
            self.water_layer_mask = numpy2tensor(water_layer_mask,dtype=torch.bool).to(device)
        else:
            self.water_layer_mask = None
            
    def _parameterization_thomson(self):
        """setting variable and gradients
        """
        # numpy2tensor
        self.vp     = numpy2tensor(self.vp   ,self.dtype).to(self.device)
        self.vs     = numpy2tensor(self.vs   ,self.dtype).to(self.device)
        self.rho    = numpy2tensor(self.rho  ,self.dtype).to(self.device)
        self.eps    = numpy2tensor(self.eps  ,self.dtype).to(self.device)
        self.gamma  = numpy2tensor(self.gamma,self.dtype).to(self.device)
        self.delta  = numpy2tensor(self.delta,self.dtype).to(self.device)
        # set model parameters
        self.vp     = torch.nn.Parameter(self.vp    ,requires_grad=self.vp_grad)
        self.vs     = torch.nn.Parameter(self.vs    ,requires_grad=self.vs_grad)
        self.rho    = torch.nn.Parameter(self.rho   ,requires_grad=self.rho_grad)
        self.eps    = torch.nn.Parameter(self.eps   ,requires_grad=False)
        self.gamma  = torch.nn.Parameter(self.gamma ,requires_grad=False)
        self.delta  = torch.nn.Parameter(self.delta ,requires_grad=False)
        return
    
    def _parameterization_Lame(self):
        """Calculate the lame parameters
        """
        mu,lamu,lam,b   = vs_vp_to_Lame(self.vp,self.vs,self.rho)
        self.mu         = mu
        self.lamu       = lamu
        self.lam        = lam
        self.b          = b
        return
    
    def _parameterization_elastic_moduli(self):
        """calculate the 21 dependent elastic moduli
        """
        # initialize elastic moduli
        CC = [C11,C12,C13,C14,C15,C16,C22,C23,C24,C25,C26,C33,C34,C35,C36,C44,C45,C46,C55,C56,C66] = elastic_moduli_init(self.nz,self.nx,self.device,self.dtype)
        # transform thomsen parameter to elastic moduli 
        C11,C13,C33,C44,C66 = thomsen_to_elastic_moduli(self.vp,self.vs,self.rho,self.eps,self.delta,self.gamma)
        CC = [C11,C12,C13,C14,C15,C16,C22,C23,C24,C25,C26,C33,C34,C35,C36,C44,C45,C46,C55,C56,C66]
        # define elastic moduli for isotropic model
        CC = [C11,C12,C13,C14,C15,C16,C22,C23,C24,C25,C26,C33,C34,C35,C36,C44,C45,C46,C55,C56,C66] = elastic_moduli_for_isotropic(CC)
        # prepare staggered grid settings
        bx,bz,muxz,C44,C55,C66 = parameter_staggered_grid(self.mu,self.b,C44,C55,C66,self.nx,self.nz)
        self.bx = bx
        self.bz = bz
        self.muxz = muxz
        CC = [C11,C12,C13,C14,C15,C16,C22,C23,C24,C25,C26,C33,C34,C35,C36,C44,C45,C46,C55,C56,C66]
        self.CC = CC
        return 
    
    def get_clone_data(self) -> Tuple:
        kwargs = super().get_clone_data()
        return kwargs
    
    def _plot_vp_vs_rho(self,**kwargs):
        """plot velocity model
        """
        plot_vp_vs_rho(self.vp,self.vs,self.rho,
                            dx=self.dx,dz=self.dz,**kwargs)
        return
    
    def _plot_eps_delta_gamma(self,**kwargs):
        """plot anisotropic parameters
        """
        plot_eps_delta_gamma(self.eps,self.delta,self.gamma,
                            dx=self.dx,dz=self.dz,**kwargs)
        return
    
    def _plot_lam_mu(self,**kwargs):
        """plot lame parameters
        """
        plot_lam_mu(self.lam,self.mu,
                            dx=self.dx,dz=self.dz,**kwargs)
        return
    
    def _plot(self,var,**kwargs):
        """plot single velocity model
        """
        model_data = self.get_model(var)
        plot_model(model_data,title=var,**kwargs)
        return
    
    def set_rho_using_empirical_function(self):
        """approximate rho via empirical relations with vp
        """
        rho         = self.rho.cpu().detach().numpy()
        vp          = self.vp.cpu().detach().numpy()
        rho_empirical  = np.power(vp, 0.25) * 310
        if self.water_layer_mask is not None:
            mask = self.water_layer_mask.cpu().detach().numpy()
            rho_empirical[mask] = rho[mask]
        rho         = numpy2tensor(rho_empirical,self.dtype).to(self.device)
        self.rho    = torch.nn.Parameter(rho   ,requires_grad=self.rho_grad)
        return
    
    def set_vp_using_empirical_function(self):
        """approximate vp via empirical relations with vs
        """
        vp = self.vp.cpu().detach().numpy()
        vs = self.vs.cpu().detach().numpy()
        vp_empirical = vs*np.sqrt(3)
        if self.water_layer_mask is not None:
            mask = self.water_layer_mask.cpu().detach().numpy()
            vp_empirical[mask] = vp[mask]
        vp = numpy2tensor(vp_empirical,self.dtype).to(self.device)
        self.vp = torch.nn.Parameter(vp,requires_grad=self.vp_grad)
        return
    
    def clip_params(self)->None:
        """Clip the model parameters to the given bounds
        """
        for par in self.pars:
            if self.lower_bound[par] is not None and self.upper_bound[par] is not None:
                # Retrieve the model parameter
                m = getattr(self, par)
                min_value = self.lower_bound[par]
                max_value = self.upper_bound[par]

                # Create a temporary copy for masking purposes
                m_temp = m.clone()  # Use .clone() instead of .copy() to avoid issues with gradients

                # Clip the values of the parameter using in-place modification with .data
                m.data.clamp_(min_value, max_value)

                # Apply the water layer mask if it is not None, using in-place modification
                if self.water_layer_mask is not None:
                    m.data = torch.where(self.water_layer_mask, m_temp.data, m.data)
        return
    

    def forward(self) -> None:
        """Forward method of the elastic model class
        
        """
        # set the constraints on the parameters if necessary
        if self.auto_update_rho:
            self.set_rho_using_empirical_function()
            
        if self.auto_update_vp:
            self.set_vp_using_empirical_function()
        
        # Clip the model parameters
        self.clip_params()
        
        # calculate the thomson/lame and elastic moduli parameters
        self._parameterization_Lame()
        self._parameterization_elastic_moduli()
        return 
    
    
class AnisotropicElasticModel(AbstractModel):
    """AnIsotropic Elastic Velocity model with parameterization of vp vs rho eps and delta. (VTI/HTI)
    """
    def __init__(self,
                ox:float,oz:float,
                nx:int,nz:int,
                dx:float,dz:float,
                vp:Optional[Union[np.array,Tensor]]         = None,     # model parameter
                vs:Optional[Union[np.array,Tensor]]         = None,
                rho:Optional[Union[np.array,Tensor]]        = None,
                eps:Optional[Union[np.array,Tensor]]        = None,
                gamma:Optional[Union[np.array,Tensor]]      = None,
                delta:Optional[Union[np.array,Tensor]]      = None,
                vp_bound: Optional[Tuple[float, float]]     = None,     # model parameter's boundary
                vs_bound: Optional[Tuple[float, float]]     = None,
                rho_bound: Optional[Tuple[float, float]]    = None,
                eps_bound: Optional[Tuple[float, float]]    = None,
                gamma_bound: Optional[Tuple[float, float]]  = None,
                delta_bound: Optional[Tuple[float, float]]  = None,
                vp_grad:Optional[bool]                      = False,    # requires gradient or not
                vs_grad:Optional[bool]                      = False,
                rho_grad:Optional[bool]                     = False,
                eps_grad:Optional[bool]                     = False,
                gamma_grad:Optional[bool]                   = False,
                delta_grad:Optional[bool]                   = False,
                free_surface:Optional[bool]                 = False,
                anisotropic_type:Optional[str]              = "vti",
                abc_type:Optional[str]                      = 'PML',
                abc_jerjan_alpha:Optional[float]            = 0.0053,
                nabc:Optional[int]                          = 20,
                auto_update_rho:Optional[bool]              = False,    # auto update parameters
                auto_update_vp:Optional[bool]               = False,
                water_layer_mask:Optional[Union[np.array,Tensor]]= None,
                device                                      = 'cpu',
                dtype                                       = torch.float32
                )->None:
        """
        Parameters:
        --------------
        ox (float), oz (float)                      : Non use, the origin coordinates of the model in the x- and z- directions (meters).
        nx (int), nz (int)                          : The number of grid points in the x- and z- directions.
        dx (float), dz (float)                      : The grid spacing in the x- and z- directions (meters).
        vp (Optional[Union[np.array, Tensor]])      : P-wave velocity model with shape (nz, nx). Default is None.
        vs (Optional[Union[np.array, Tensor]])      : S-wave velocity model with shape (nz, nx). Default is None.
        rho (Optional[Union[np.array, Tensor]])     : Density model with shape (nz, nx). Default is None.
        eps (Optional[Union[np.array, Tensor]])     : Anisotropic parameter epsilon (vti/hti model), shape (nz, nx). Default is None.
        gamma (Optional[Union[np.array, Tensor]])   : Anisotropic parameter gamma (vti/hti model), shape (nz, nx). Default is None.
        delta (Optional[Union[np.array, Tensor]])   : Anisotropic parameter delta (vti/hti model), shape (nz, nx). Default is None.
        vp_bound (Optional[Tuple[float, float]])    : The lower and upper bounds for the P-wave velocity model. Default is None.
        vs_bound (Optional[Tuple[float, float]])    : The lower and upper bounds for the S-wave velocity model. Default is None.
        rho_bound (Optional[Tuple[float, float]])   : The lower and upper bounds for the density model. Default is None.
        eps_bound (Optional[Tuple[float, float]])   : The lower and upper bounds for epsilon. Default is None.
        gamma_bound (Optional[Tuple[float, float]]) : The lower and upper bounds for gamma. Default is None.
        delta_bound (Optional[Tuple[float, float]]) : The lower and upper bounds for delta. Default is None.
        vp_grad (Optional[bool])                    : Whether to compute the gradient of P-wave velocity. Default is False.
        vs_grad (Optional[bool])                    : Whether to compute the gradient of S-wave velocity. Default is False.
        rho_grad (Optional[bool])                   : Whether to compute the gradient of the density model. Default is False.
        eps_grad (Optional[bool])                   : Whether to compute the gradient of epsilon. Default is False.
        gamma_grad (Optional[bool])                 : Whether to compute the gradient of gamma. Default is False.
        delta_grad (Optional[bool])                 : Whether to compute the gradient of delta. Default is False.
        free_surface (Optional[bool])               : Whether to include a free surface in the model. Default is False.
        anisotropic_type (Optional[str])            : Type of anisotropic model ('vti', 'hti', etc.). Default is 'vti'.
        abc_type (Optional[str])                    : Type of absorbing boundary condition ('PML', 'Jerjan', etc.). Default is 'PML'.
        abc_jerjan_alpha (Optional[float])          : Attenuation factor for Jerjan boundary condition. Default is 0.0053.
        nabc (Optional[int])                        : Number of absorbing boundary cells. Default is 20.
        auto_update_rho (Optional[bool])            : Whether to auto-update the density model during inversion. Default is False.
        auto_update_vp (Optional[bool])             : Whether to auto-update the P-wave velocity model during inversion. Default is False.
        water_layer_mask (Optional[Union[np.array, Tensor]]) : Mask for the water layer (not update), if applicable. Default is None.
        device (str)                                : Device for running the model ('cpu' or 'cuda'). Default is 'cpu'.
        dtype (torch.dtype)                         : Data type for PyTorch tensors. Default is torch.float32.
        """
        # initialize the common model parameters
        super().__init__(ox,oz,nx,nz,dx,dz,free_surface,abc_type,abc_jerjan_alpha,nabc,device,dtype)

        # initialize the thomson model model parameters
        self.pars       = ["vp","vs","rho","eps","gamma","delta"]
        self.vp         = vp.copy()
        self.vs         = vs.copy()
        self.rho        = rho.copy()
        self.eps        = eps.copy()
        self.gamma      = gamma.copy()
        self.delta      = delta.copy()
        self.vp_grad    = vp_grad
        self.vs_grad    = vs_grad
        self.rho_grad   = rho_grad
        self.eps_grad   = eps_grad
        self.gamma_grad = gamma_grad
        self.delta_grad = delta_grad
        
        self.anisotropic_type = anisotropic_type
        self._parameterization_thomson()
        
        # initialize the lame constant
        self.mu         = None
        self.lamu       = None
        self.lam        = None
        self.muxz       = None
        self.b          = None
        self.bx         = None
        self.bz         = None
        self._parameterization_Lame()
        
        # initialize the elastic moduli
        self.CC         = []
        self._parameterization_elastic_moduli()
        
        # set model bounds
        self.lower_bound["vp"]      = vp_bound[0] if vp_bound  is not None else None
        self.lower_bound["vs"]      = vs_bound[0] if vs_bound  is not None else None
        self.lower_bound["rho"]     = rho_bound[0] if rho_bound is not None else None
        self.lower_bound['eps']     = eps_bound[0] if eps_bound is not None else None
        self.lower_bound['gamma']   = gamma_bound[0] if gamma_bound is not None else None
        self.lower_bound['delta']   = delta_bound[0] if delta_bound is not None else None
        self.upper_bound["vp"]      = vp_bound[1] if vp_bound  is not None else None
        self.upper_bound["vs"]      = vs_bound[1] if vs_bound  is not None else None
        self.upper_bound["rho"]     = rho_bound[1] if rho_bound is not None else None
        self.upper_bound["eps"]     = eps_bound[1] if eps_bound  is not None else None
        self.upper_bound["gamma"]   = gamma_bound[1] if gamma_bound  is not None else None
        self.upper_bound["delta"]   = delta_bound[1] if delta_bound is not None else None
        
        # set model gradients
        self.requires_grad["vp"]    = self.vp_grad
        self.requires_grad["vs"]    = self.vs_grad
        self.requires_grad["rho"]   = self.rho_grad
        self.requires_grad["eps"]   = self.eps_grad
        self.requires_grad["gamma"] = self.gamma_grad
        self.requires_grad["delta"] = self.delta_grad
        
        # check the input model
        self._check_bounds()
        self.check_dims()
        
        # update rho using the empirical function
        self.auto_update_rho = auto_update_rho
        self.auto_update_vp  = auto_update_vp
        
        if water_layer_mask is not None:
            self.water_layer_mask = numpy2tensor(water_layer_mask,dtype=torch.bool).to(device)
        else:
            self.water_layer_mask = None
                
    def _parameterization_thomson(self):
        # numpy2tensor
        self.vp     = numpy2tensor(self.vp   ,self.dtype).to(self.device)
        self.vs     = numpy2tensor(self.vs   ,self.dtype).to(self.device)
        self.rho    = numpy2tensor(self.rho  ,self.dtype).to(self.device)
        self.eps    = numpy2tensor(self.eps  ,self.dtype).to(self.device)
        self.gamma  = numpy2tensor(self.gamma,self.dtype).to(self.device)
        self.delta  = numpy2tensor(self.delta,self.dtype).to(self.device)
        # set model parameters
        self.vp     = torch.nn.Parameter(self.vp    ,requires_grad=self.vp_grad)
        self.vs     = torch.nn.Parameter(self.vs    ,requires_grad=self.vs_grad)
        self.rho    = torch.nn.Parameter(self.rho   ,requires_grad=self.rho_grad)
        self.eps    = torch.nn.Parameter(self.eps   ,requires_grad=self.eps_grad)
        self.gamma  = torch.nn.Parameter(self.gamma ,requires_grad=self.gamma_grad)
        self.delta  = torch.nn.Parameter(self.delta ,requires_grad=self.delta_grad)
        return
    
    def _parameterization_Lame(self):
        mu,lamu,lam,b   = vs_vp_to_Lame(self.vp,self.vs,self.rho)
        self.mu         = mu
        self.lamu       = lamu
        self.lam        = lam
        self.b          = b
        return
    
    def _parameterization_elastic_moduli(self):
        # initialize elastic moduli
        CC = [C11,C12,C13,C14,C15,C16,C22,C23,C24,C25,C26,C33,C34,C35,C36,C44,C45,C46,C55,C56,C66] = elastic_moduli_init(self.nz,self.nx,self.device,self.dtype)
        # transform thomsen parameter to elastic moduli 
        C11,C13,C33,C44,C66 = thomsen_to_elastic_moduli(self.vp,self.vs,self.rho,self.eps,self.delta,self.gamma)
        CC = [C11,C12,C13,C14,C15,C16,C22,C23,C24,C25,C26,C33,C34,C35,C36,C44,C45,C46,C55,C56,C66]
        # define elastic moduli for anisotropic model
        CC = [C11,C12,C13,C14,C15,C16,C22,C23,C24,C25,C26,C33,C34,C35,C36,C44,C45,C46,C55,C56,C66] = elastic_moduli_for_TI(CC,anisotropic_type=self.anisotropic_type)
        # prepare staggered grid settings
        bx,bz,muxz,C44,C55,C66 = parameter_staggered_grid(self.mu,self.b,C44,C55,C66,self.nx,self.nz)
        self.bx = bx
        self.bz = bz
        self.muxz = muxz
        CC = [C11,C12,C13,C14,C15,C16,C22,C23,C24,C25,C26,C33,C34,C35,C36,C44,C45,C46,C55,C56,C66]
        self.CC = CC
        return 
    
    def _plot_vp_vs_rho(self,**kwargs):
        """plot velocity model
        """
        plot_vp_vs_rho(self.vp,self.vs,self.rho,
                            dx=self.dx,dz=self.dz,**kwargs)
        return
    
    def _plot_eps_delta_gamma(self,**kwargs):
        """plot anisotropic parameters
        """
        plot_eps_delta_gamma(self.eps,self.delta,self.gamma,
                            dx=self.dx,dz=self.dz,**kwargs)
        return
    
    def _plot_lam_mu(self,**kwargs):
        """plot lame parameters
        """
        plot_lam_mu(self.lam,self.mu,
                            dx=self.dx,dz=self.dz,**kwargs)
        return
    
    def _plot(self,var,**kwargs):
        """plot single velocity model
        """
        model_data = self.get_model(var)
        plot_model(model_data,title=var,**kwargs)
        return
    
    def set_rho_using_empirical_function(self):
        """approximate rho via empirical relations with vp
        """
        vp          = self.vp.cpu().detach().numpy()
        rho         = np.power(vp, 0.25) * 310
        rho         = numpy2tensor(rho,self.dtype).to(self.device)
        self.rho    = torch.nn.Parameter(rho   ,requires_grad=self.rho_grad)
        return
    
    def set_vp_using_empirical_function(self):
        """approximate vp via empirical relations with vs
        """
        vs = self.vs.cpu().detach().numpy()
        vp = vs*np.sqrt(3)
        vp = numpy2tensor(vp,self.dtype).to(self.device)
        self.vp = torch.nn.Parameter(vp,requires_grad=self.vp_grad)
        return
    
    def clip_params(self)->None:
        """Clip the model parameters to the given bounds
        """
        for par in self.pars:
            if self.lower_bound[par] is not None and self.upper_bound[par] is not None:
                # Retrieve the model parameter
                m = getattr(self, par)
                min_value = self.lower_bound[par]
                max_value = self.upper_bound[par]

                # Create a temporary copy for masking purposes
                m_temp = m.clone()  # Use .clone() instead of .copy() to avoid issues with gradients

                # Clip the values of the parameter using in-place modification with .data
                m.data.clamp_(min_value, max_value)

                # Apply the water layer mask if it is not None, using in-place modification
                if self.water_layer_mask is not None:
                    m.data = torch.where(self.water_layer_mask, m_temp.data, m.data)
        return
    
        
    def forward(self) -> None:
        """Forward method of the elastic model class
        """
        # set the constraints on the parameters if necessary
        if self.auto_update_rho:
            self.set_rho_using_empirical_function()
            
        if self.auto_update_vp:
            self.set_vp_using_empirical_function()
            
        # Clip the model parameters
        self.clip_params()
        
        # calculate the thomson/lame and elastic moduli parameters
        self._parameterization_Lame()
        self._parameterization_elastic_moduli()
        return 


'''
* Author: LiuFeng(SJTU) : liufeng2317@sjtu.edu.cn
* Date: 2024-04-20 09:32:43
* LastEditors: LiuFeng
* LastEditTime: 2024-05-15 19:30:38
* Description: 
* Copyright (c) 2024 by liufeng, Email: liufeng2317@sjtu.edu.cn, All Rights Reserved.
'''

import numpy as np
import torch
from .base import Misfit
from typing import Optional

"""
    Wasserstein distance (W1 Loss)
        https://github.com/YangFangShu/FWIGAN/blob/main/FWIGAN/utils/misfit.py
"""

def transform(f, g, trans_type, theta):
    """
    Apply a transformation to make f and g non-negative.
    
    Args:
        f, g (Tensor): Seismic data, shape [num_time_steps, num_shots*num_receivers_per_shot].
        trans_type (str): Transformation type ('linear', 'abs', 'square', 'exp', 'softplus').
        theta (float): Scalar parameter for transformation.
        
    Returns:
        mu (Tensor): Transformed f.
        nu (Tensor): Transformed g.
        d (Tensor): Derivative of transformed f for potential use in gradient-based methods.
    """
    c = 0.0  # Initialize the offset constant
    device = f.device
    if trans_type == 'linear':
        # Linear transformation with offset
        min_value = torch.min(f.detach().min(), g.detach().min())
        mu, nu = f, g
        c = -min_value if min_value < 0 else 0
        c = c * theta  # Scale by theta for flexibility
        d = torch.ones(f.shape).to(device)
    elif trans_type == 'abs':
        # Absolute value transformation
        mu, nu = torch.abs(f), torch.abs(g)
        d = torch.sign(f).to(device)
    elif trans_type == 'square':
        # Squaring transformation
        mu = f * f
        nu = g * g
        d = 2 * f
    elif trans_type == 'exp':
        # Exponential transformation, scaled by theta
        mu = torch.exp(theta * f)
        nu = torch.exp(theta * g)
        d = theta * mu
    elif trans_type == 'softplus':
        # Softplus transformation for smooth non-negativity
        mu = torch.log(torch.exp(theta * f) + 1)
        nu = torch.log(torch.exp(theta * g) + 1)
        d = theta / (torch.exp(-theta * f) + 1e-18)  # Avoid division by zero
    else:
        mu, nu = f, g
        d = torch.ones(f.shape).to(device)
    # Ensure positive values for mu and nu by adding a small constant
    mu = mu + c + 1e-18
    nu = nu + c + 1e-18
    return mu, nu, d


def trace_sum_normalize(x, time_dim=0):
    """
    Normalize each trace by its sum along the specified dimension.
    
    Args:
        x (Tensor): Input tensor.
        time_dim (int): Dimension for time steps.
        
    Returns:
        Tensor: Normalized tensor.
    """
    x = x / (x.sum(dim=time_dim, keepdim=True) + 1e-18)  # Avoid division by zero
    return x

def trace_max_normalize(x, time_dim=0):
    """
    normalization with the maximum value of each trace (the value of each trace is in [-1,1] after the processing)
    note that the channel should be 1
    """
    x_max,_ = torch.max(x.abs(), dim=time_dim, keepdim=True)
    x = x / (x_max+1e-18)
    return x

class Misfit_Wasserstein1(Misfit):
    def __init__(self,p=1,trans_type='linear',theta=1,dt=1) -> None:
        """
            trans_type: linear, abs, square, exp, softplus
        """
        super().__init__()
        self.p = 1
        self.trans_type = trans_type
        self.theta = theta
        self.dt = dt
        
    def forward(self,syn:torch.Tensor,obs:torch.Tensor):
        assert syn.shape == obs.shape
        
        # Flatten the input tensors for transformation (shape: [num_shots, num_time_steps * num_receivers])
        num_shots, num_time_steps, num_receivers = syn.shape
        
        # [num_shots, num_time_steps, num_receivers] --> [num_shots, num_receivers, num_time_steps]
        obs_transposed = obs.permute(1, 0, 2) 
        syn_transposed = syn.permute(1, 0, 2)
        
        # Reshape input tensors for transformation
        syn_flat = syn_transposed.reshape(num_time_steps, num_shots * num_receivers)
        obs_flat = obs_transposed.reshape(num_time_steps, num_shots * num_receivers)
        
        # Transform signals to ensure non-negativity
        mu, nu, d = transform(syn_flat, obs_flat, self.trans_type, self.theta)
        
        assert mu.min() > 0
        assert nu.min() > 0
        
        # Normalize each trace by its sum
        mu = trace_sum_normalize(mu, time_dim=0)
        nu = trace_sum_normalize(nu, time_dim=0)

        # Compute cumulative sums over the time dimension
        F = torch.cumsum(mu, dim=0)  # Keep the cumulative sum along the flattened time dimension
        G = torch.cumsum(nu, dim=0)  # Keep the cumulative sum along the flattened time dimension
        
        w1loss = (torch.abs(F-G) ** self.p).sum()*self.dt
        return w1loss


'''
* Author: LiuFeng(SJTU) : liufeng2317@sjtu.edu.cn
* Date: 2024-04-20 09:32:43
* LastEditors: LiuFeng
* LastEditTime: 2024-05-15 19:30:38
* Description: 
* Copyright (c) 2024 by liufeng, Email: liufeng2317@sjtu.edu.cn, All Rights Reserved.
'''

from .base import Misfit
import torch

class Misfit_waveform_L2(Misfit):
    '''Waveform L2-norm difference misfit (Tarantola, 1984)
    
    Parameters:
    -----------
        dt (float)      : Time sampling interval.
        obs (Tensor)    : Observed waveform.
        syn (Tensor)    : Synthetic waveform.
    '''
    def __init__(self, dt=1) -> None:
        super().__init__()
        self.dt = dt
    
    def forward(self, obs: torch.Tensor, syn: torch.Tensor) -> torch.Tensor:
        '''Compute the L2-norm waveform misfit between observed and synthetic data.

        Args:
            obs (Tensor): Observed waveform.
            syn (Tensor): Synthetic waveform.
        
        Returns:
            Tensor: L2-norm misfit loss.
        '''
        mask1    = torch.sum(torch.abs(obs),axis=1) == 0
        mask2    = torch.sum(torch.abs(syn),axis=1) == 0
        mask     = ~(mask1 * mask2)
        
        # Calculate residuals by subtracting synthetic from observed data
        rsd = obs - syn

        # Compute the L2-norm loss as the square root of the sum of squared residuals, weighted by dt
        # Summation along axis=1 (channels) for each sample, then take square root and sum over all samples
        loss = torch.sum(torch.sqrt(torch.sum(rsd * rsd * self.dt, axis=1)[mask]))
        
        return loss



'''
* Author: LiuFeng(USTC) : liufeng2317@sjtu.edu.cn
* Date: 2024-05-05 19:51:52
* LastEditors: LiuFeng
* LastEditTime: 2024-05-05 20:29:15
* Description: 
* Copyright (c) 2024 by liufeng, Email: liufeng2317@sjtu.edu.cn, All Rights Reserved.
'''
# https://pytorch.org/docs/stable/optim.html
from torch.optim import Adadelta,Adagrad,Adam,AdamW,Adamax,ASGD,NAdam,RAdam,Rprop,SGD

from ncg_optimizer import LCG, BASIC as NLCG

from .optimizer_swit import Swit_Optimizer



'''
* Author: LiuFeng(SJTU) : liufeng2317@sjtu.edu.cn
* Date: 2023-12-10 20:30:00
* LastEditors: LiuFeng
* LastEditTime: 2024-05-20 11:22:22
* FilePath: /ADFWI/TorchInversion/demo.py
* Description: 
* Copyright (c) 2023 by ${liufeng2317} email: ${liufeng2317}, All Rights Reserved.
'''

import numpy as np 
import matplotlib.pyplot as plt
import os 
import obspy
from scipy.interpolate import interp2d
from scipy.ndimage import gaussian_filter
from scipy.ndimage import gaussian_filter1d


############################################################
#                   Marmousi Model
############################################################

def download_marmousi_model(in_dir):
    # Check for model files and download if missing
    for filename in ["vp_marmousi-ii.segy.gz", "vs_marmousi-ii.segy.gz", "density_marmousi-ii.segy.gz"]:
        if not os.path.exists(os.path.join(in_dir, filename)):
            os.system(f"wget http://www.agl.uh.edu/downloads/{filename} -P {in_dir}")
    return

def load_marmousi_model(in_dir):
    """Load the Marmousi model data from the specified directory.

    Args:
        in_dir (str): The directory where the Marmousi model files are stored.

    Returns:
        dict: A dictionary containing the velocity and density models 
              along with spatial coordinates and increments.
    """
    download_marmousi_model(in_dir)
    
    def extract_data(meta):
        data = []
        for trace in meta:
            data.append(trace.data)
        return np.array(data)

    # Read data from SEGY files and convert units
    vs  = extract_data(obspy.read(os.path.join(in_dir, "vs_marmousi-ii.segy.gz"), format='segy')) * 1e3
    vp  = extract_data(obspy.read(os.path.join(in_dir, "vp_marmousi-ii.segy.gz"), format='segy')) * 1e3
    rho = extract_data(obspy.read(os.path.join(in_dir, "density_marmousi-ii.segy.gz"), format='segy')) * 1e3

    # Define spatial ranges and create coordinate arrays
    x_range = [0, 17000]  # in meters
    y_range = [0, 3500]   # in meters
    nx, ny = vp.shape
    x = np.linspace(x_range[0], x_range[1], nx)
    y = np.linspace(y_range[0], y_range[1], ny)

    # Create a dictionary to hold the Marmousi model data
    marmousi_model = {
        'vp': vp,
        'vs': vs,
        'rho': rho,
        'x': x,
        'y': y,
        'dx': x[1] - x[0],
        'dy': y[1] - y[0]
    }
    
    return marmousi_model  # Return the complete model data


def resample_marmousi_model(x, y, model):
    """Resample the Marmousi model to a new grid defined by x and y.

    Args:
        x (np.ndarray): The new x coordinates for resampling.
        y (np.ndarray): The new y coordinates for resampling.
        model (dict): The original Marmousi model containing velocity and density data.

    Returns:
        dict: A new model dictionary with resampled velocities and density along with new coordinates.
    """

    # Perform cubic interpolation for shear wave velocity (vs)
    vs = interp2d(model['y'], model['x'], model['vs'], kind='cubic')(y, x)
    
    # Perform cubic interpolation for primary wave velocity (vp)
    vp = interp2d(model['y'], model['x'], model['vp'], kind='cubic')(y, x)
    
    # Perform cubic interpolation for density (rho)
    rho = interp2d(model['y'], model['x'], model['rho'], kind='cubic')(y, x)

    # Create a new dictionary to hold the resampled model data
    new_model = {
        'vp': vp,
        'vs': vs,
        'rho': rho,
        'x': x,
        'y': y,
        'dx': x[1] - x[0],
        'dy': y[1] - y[0]
    }
    
    return new_model


def get_smooth_marmousi_model(model, gaussian_kernel=10, mask_extra_detph=2, rcv_depth=10):
    """Smooth the Marmousi model using a Gaussian filter.

    Args:
        model (dict): The original Marmousi model containing velocity and density data.
        gaussian_kernel (int): Standard deviation for the Gaussian kernel.
        mask_extra_detph (int): Additional depth levels to mask during smoothing.
        rcv_depth (int): Depth of the receiver.

    Returns:
        dict: A new model dictionary with smoothed velocities and density.
    """

    # Create copies of the velocity and density models for smoothing
    vp = model['vp'].copy()
    vs = model['vs'].copy()
    rho = model['rho'].copy()

    if mask_extra_detph > 0:
        # Smooth from a specified depth downwards
        vp[:, rcv_depth + mask_extra_detph:]  = gaussian_filter(vp[:, rcv_depth + mask_extra_detph:], 
                                                               [gaussian_kernel, gaussian_kernel], mode='reflect')
        vs[:, rcv_depth + mask_extra_detph:]  = gaussian_filter(vs[:, rcv_depth + mask_extra_detph:], 
                                                               [gaussian_kernel, gaussian_kernel], mode='reflect')
        rho[:, rcv_depth + mask_extra_detph:] = gaussian_filter(rho[:, rcv_depth + mask_extra_detph:], 
                                                                [gaussian_kernel, gaussian_kernel], mode='reflect')
    else:
        # Smooth the entire model
        vp  = gaussian_filter(vp, [gaussian_kernel, gaussian_kernel], mode='reflect')
        vs  = gaussian_filter(vs, [gaussian_kernel, gaussian_kernel], mode='reflect')
        rho = gaussian_filter(rho, [gaussian_kernel, gaussian_kernel], mode='reflect')

    # Create a new dictionary for the smoothed model data
    new_model = {
        'vp': vp,
        'vs': vs,
        'rho': rho,
        'x': model['x'],
        'y': model['y'],
        'dx': model['dx'],
        'dy': model['dy']
    }
    
    return new_model



def get_linear_vel_model(model, vp_min=None, vp_max=None, vs_min=None, vs_max=None, mask_depth=10):
    """Generate a linear velocity model based on the input Marmousi model.

    Args:
        model (dict): The original Marmousi model containing velocity and density data.
        vp_min (float, optional): Minimum value for the primary wave velocity.
        vp_max (float, optional): Maximum value for the primary wave velocity.
        vs_min (float, optional): Minimum value for the shear wave velocity.
        vs_max (float, optional): Maximum value for the shear wave velocity.

    Returns:
        dict: A new model dictionary with linearly varying velocities and original density.
    """
    
    vp_true  = np.array(model['vp']).T
    vs_true  = np.array(model['vs']).T
    rho_true = np.array(model['rho']).T
    nz, nx = vp_true.shape
    vp = np.ones_like(vp_true)
    vs = np.ones_like(vs_true)

    vp[:mask_depth, :] = vp_true[:mask_depth, :]
    vs[:mask_depth, :] = vs_true[:mask_depth, :]
    
    # Determine velocity limits if not provided
    if vp_min is None and vp_max is None:
        vp_min, vp_max = np.min(vp_true[mask_depth:, :]), np.max(vp_true[mask_depth:, :])
    if vs_min is None and vs_max is None:
        vs_min, vs_max = np.min(vs_true[mask_depth:, :]), np.max(vs_true[mask_depth:, :])

    # Create linearly spaced values for velocities below the mask depth
    vp_line = np.linspace(vp_min, vp_max, nz - mask_depth).reshape(-1, 1)
    vs_line = np.linspace(vs_min, vs_max, nz - mask_depth).reshape(-1, 1)
    vp[mask_depth:, :] *= vp_line
    vs[mask_depth:, :] *= vs_line
    rho = rho_true

    # Create a new dictionary to hold the new model data
    new_model = {
        'vp': vp.T,
        'vs': vs.T,
        'rho': rho.T,
        'x': model['x'],
        'y': model['y'],
        'dx': model['dx'],
        'dy': model['dy']
    }
    
    return new_model


def get_linear_marmousi2_model(model,smooth_kernel=10,idx=None,mask_depth=10):
    """Generate a linear velocity model based on the input Marmousi model.

    Args:
        model (dict): The original Marmousi model containing velocity and density data.
        vp_min (float, optional): Minimum value for the primary wave velocity.
        vp_max (float, optional): Maximum value for the primary wave velocity.
        vs_min (float, optional): Minimum value for the shear wave velocity.
        vs_max (float, optional): Maximum value for the shear wave velocity.

    Returns:
        dict: A new model dictionary with linearly varying velocities and original density.
    """
    
    vp_true  = np.array(model['vp']).T
    vs_true  = np.array(model['vs']).T
    rho_true = np.array(model['rho']).T
    nz, nx = vp_true.shape
    vp = np.ones_like(vp_true)
    vs = np.ones_like(vs_true)

    vp[:mask_depth, :] = vp_true[:mask_depth, :]
    vs[:mask_depth, :] = vs_true[:mask_depth, :]
    
    if idx is None:
        vp      = np.ones_like(vp_true)*np.mean(vp_true,axis=1).reshape(-1,1)
        vs      = np.ones_like(vp_true)*np.mean(vs_true,axis=1).reshape(-1,1)
        rho     = np.ones_like(rho_true)*np.mean(rho_true,axis=1).reshape(-1,1)
    else:
        vp      = np.ones_like(vp_true)*vp_true[:,idx].reshape(-1,1)    
        vs      = np.ones_like(vs_true)*vs_true[:,idx].reshape(-1,1)    
        rho     = np.ones_like(rho_true)*rho_true[:,idx].reshape(-1,1)      
        
    vp      = gaussian_filter1d(vp      , sigma=smooth_kernel, axis=0)
    vs      = gaussian_filter1d(vs      , sigma=smooth_kernel, axis=0)
    rho     = gaussian_filter1d(rho     , sigma=smooth_kernel, axis=0)
    
    # Create a new dictionary to hold the new model data
    new_model = {
        'vp': vp.T,
        'vs': vs.T,
        'rho': rho.T,
        'x': model['x'],
        'y': model['y'],
        'dx': model['dx'],
        'dy': model['dy']
    }
    
    return new_model

############################################################
#                   Layer Model
############################################################
from scipy.interpolate import interp1d

def step_profile_layerModel(x_range, y_range, step):
    """Generate a step-profile layer model based on specified ranges and step size.

    Args:
        x_range (list): The x-coordinate range (not used in calculations).
        y_range (list): The y-coordinate range defining the depth limits.
        step (float): The step size for the profile.

    Returns:
        tuple: Arrays representing y-coordinates, primary wave velocity (vp),
               shear wave velocity (vs), and density (rho) for the model.
    """

    # Create rounded y-coordinates based on the specified range and step size
    y_step1 = np.round(np.arange(y_range[0], y_range[1] + step, step) / step) * step
    
    # Calculate velocities and density linearly based on y-coordinates
    vp_step1 = y_step1 / (y_range[1] - y_range[0]) * (6.5 - 5) + 3
    vs_step1 = y_step1 / (y_range[1] - y_range[0]) * (4.48 - 3.46) + 2.46
    rho_step1 = y_step1 / (y_range[1] - y_range[0]) * (3.32 - 2.72) + 2.72

    # Create second set of y-coordinates shifted by a small amount
    y_step2 = y_step1 + (y_step1[1] - y_step1[0] - step / 5)
    
    # Combine and sort the y-coordinates and corresponding properties
    idy = np.argsort(np.hstack([y_step1, y_step2]))
    y_step = np.hstack([y_step1, y_step2])[idy]
    vp_step = np.hstack([vp_step1, vp_step1])[idy]
    vs_step = np.hstack([vs_step1, vs_step1])[idy]
    rho_step = np.hstack([rho_step1, rho_step1])[idy]

    # Set the last entry to the second-to-last value to avoid discontinuity
    vp_step[-1:] = vp_step[-2]
    vs_step[-1:] = vs_step[-2]
    rho_step[-1:] = rho_step[-2]
    
    return y_step, vp_step, vs_step, rho_step


def build_layer_model(x, y, step):
    """Construct a layered model using specified x and y coordinates and a step size.

    Args:
        x (array): Array of x-coordinates defining the spatial extent.
        y (array): Array of y-coordinates defining the depth levels.
        step (float): The step size for the layer profile.

    Returns:
        dict: A dictionary containing the modeled properties: 
              primary wave velocity (vp), shear wave velocity (vs),
              density (rho), and the coordinate information.
    """

    # Generate step-profile layer model based on the specified x and y ranges
    y_step, vp_step, vs_step, rho_step = step_profile_layerModel([x[0], x[-1]], [y[0], y[-1]], step)

    # Interpolate velocities and density to match the specified y-coordinates
    vp = interp1d(y_step, vp_step, kind='slinear')(y)
    vs = interp1d(y_step, vs_step, kind='slinear')(y)
    rho = interp1d(y_step, rho_step, kind='slinear')(y)

    # Expand the velocity and density arrays to match the x-dimension
    vp = np.tile(vp[np.newaxis, :], [len(x), 1])
    vs = np.tile(vs[np.newaxis, :], [len(x), 1])
    rho = np.tile(rho[np.newaxis, :], [len(x), 1])

    # Create a model dictionary to hold the properties and coordinates
    model = {
        'vp': vp,
        'vs': vs,
        'rho': rho,
        'x': x,
        'y': y,
        'dx': x[1] - x[0],
        'dy': y[1] - y[0]
    }

    return model


def get_smooth_layer_model(model, smooth_kernel=10):
    """Apply Gaussian smoothing to the velocity and density fields of the input model.

    Args:
        model (dict): A dictionary containing the properties of the model 
                       (vp, vs, rho, x, y, dx, dy).
        smooth_kernel (int, optional): The standard deviation for Gaussian kernel smoothing. 
                                        Defaults to 10.

    Returns:
        dict: A new model dictionary with smoothed properties.
    """

    # Apply Gaussian smoothing to the velocity and density fields
    vp = gaussian_filter(model['vp'].copy(), [smooth_kernel, smooth_kernel], mode='reflect')
    vs = gaussian_filter(model['vs'].copy(), [smooth_kernel, smooth_kernel], mode='reflect')
    rho = gaussian_filter(model['rho'].copy(), [smooth_kernel, smooth_kernel], mode='reflect')

    # Prepare a new model dictionary with smoothed properties
    new_model = {
        'vp': vp,
        'vs': vs,
        'rho': rho,
        'x': model['x'],
        'y': model['y'],
        'dx': model['dx'],
        'dy': model['dy']
    }

    return new_model


############################################################
#                   Layer Anomaly Model
############################################################
from scipy.interpolate import interp1d
def step_profile_anomaly(x_range, y_range, step):
    """Generate a stepped profile with anomalies in velocity and density.

    Args:
        x_range (list): The range of x-coordinates (not used in this function).
        y_range (list): The range of y-coordinates defining the depth.
        step (float): The step size for creating the profile.

    Returns:
        tuple: A tuple containing arrays of depth (y_step), 
               primary wave velocity (vp_step), shear wave velocity (vs_step), 
               and density (rho_step).
    """

    # Create the first step profile based on the y_range and step
    y_step1 = np.round(np.arange(y_range[0], y_range[1] + step, step) / step) * step
    
    # Calculate properties for the first step profile
    vp_step1 = y_step1 / (y_range[1] - y_range[0]) * (8.04 - 5.8) + 5.8  # Primary wave velocity
    vs_step1 = y_step1 / (y_range[1] - y_range[0]) * (4.48 - 3.46) + 3.46  # Shear wave velocity
    rho_step1 = y_step1 / (y_range[1] - y_range[0]) * (3.32 - 2.72) + 2.72  # Density

    # Create the second step profile with an adjusted depth
    y_step2 = y_step1 + (y_step1[1] - y_step1[0] - 1)
    vp_step2 = vp_step1  # Keep primary wave velocity same for second step
    vs_step2 = vs_step1  # Keep shear wave velocity same for second step
    rho_step2 = rho_step1  # Keep density same for second step

    # Combine the two step profiles and sort them
    combined_y = np.hstack([y_step1, y_step2])
    combined_vp = np.hstack([vp_step1, vp_step2])
    combined_vs = np.hstack([vs_step1, vs_step2])
    combined_rho = np.hstack([rho_step1, rho_step2])

    # Sort the combined arrays based on depth
    idy = np.argsort(combined_y)  # Sort indices
    y_step = combined_y[idy]  # Combined and sorted depth
    vp_step = combined_vp[idy]  # Combined and sorted primary wave velocity
    vs_step = combined_vs[idy]  # Combined and sorted shear wave velocity
    rho_step = combined_rho[idy]  # Combined and sorted density

    # Ensure the last values are consistent with the second-to-last values
    vp_step[-1] = vp_step[-2]  # Set last vp value to second-to-last
    vs_step[-1] = vs_step[-2]  # Set last vs value to second-to-last
    rho_step[-1] = rho_step[-2]  # Set last rho value to second-to-last
    
    return y_step, vp_step, vs_step, rho_step  # Return the generated profiles


def build_anomaly_background_model(x, y, step):
    """Construct a model with an anomaly background based on step profiles.

    Args:
        x (numpy.ndarray): The x-coordinates for the model.
        y (numpy.ndarray): The y-coordinates (depths) for the model.
        step (float): The step size used to generate the profile.

    Returns:
        dict: A dictionary containing the velocity and density models along with coordinate information.
    """

    # Generate stepped profile with anomalies
    y_step, vp_step, vs_step, rho_step = step_profile_anomaly([x[0], x[-1]], [y[0], y[-1]], step)

    # Interpolate velocities and density values based on the step profile
    vp = interp1d(y_step, vp_step, kind='slinear')(y)  # Interpolated primary wave velocity
    vs = interp1d(y_step, vs_step, kind='slinear')(y)  # Interpolated shear wave velocity
    rho = interp1d(y_step, rho_step, kind='slinear')(y)  # Interpolated density

    # Create 2D arrays for the model parameters by tiling the interpolated values
    vp = np.tile(vp[np.newaxis, :], [len(x), 1])  # Expand vp to 2D
    vs = np.tile(vs[np.newaxis, :], [len(x), 1])  # Expand vs to 2D
    rho = np.tile(rho[np.newaxis, :], [len(x), 1])  # Expand rho to 2D
    
    # Prepare the model dictionary with relevant data
    model = {
        'vp': vp,  # Primary wave velocity
        'vs': vs,  # Shear wave velocity
        'rho': rho,  # Density
        'x': x,  # X-coordinates
        'y': y,  # Y-coordinates (depths)
        'dx': x[1] - x[0],  # X-coordinates spacing
        'dy': y[1] - y[0]   # Y-coordinates spacing
    }
    
    return model  # Return the constructed model



def get_anomaly_model(layer_model, n_pml):
    """Generate a model with anomalies based on the given layer model.

    Args:
        layer_model (dict): The base model containing velocity and density.
        n_pml (int): The number of PML (Perfectly Matched Layer) cells.

    Returns:
        dict: A dictionary containing the modified model with anomalies.
    """
    
    # Extracting coordinates and model parameters
    x = layer_model['x']
    y = layer_model['y']
    dx = x[1] - x[0]
    dy = y[1] - y[0]
    vp = layer_model['vp'].copy()
    vs = layer_model['vs'].copy()
    rho = layer_model['rho'].copy()
    
    # Define anomaly 1 parameters
    x0 = (x[-1] - 1 * n_pml * dx) * 2 / 3 + 0.5 * n_pml * dx
    y0 = (y[-1] - 1 * n_pml * dy) * 1 / 3 + 0.5 * n_pml * dy
    a = x[-1] / 6
    b = y[-1] / 10
    anomaly1 = np.zeros_like(vp)
    
    # Create first anomaly
    for i, xi in enumerate(x):
        for j, yj in enumerate(y):
            if ((xi - x0) / a) ** 2 + ((yj - y0) / b) ** 2 < 1:
                anomaly1[i, j] = 1
    
    # Define anomaly 2 parameters
    x0 = (x[-1] - 1 * n_pml * dx) / 3 + 0.5 * n_pml * dx
    y0 = (y[-1] - 1 * n_pml * dy) * 2 / 3 + 0.5 * n_pml * dy
    anomaly2 = np.zeros_like(vp)

    # Create second anomaly
    for i, xi in enumerate(x):
        for j, yj in enumerate(y):
            if ((xi - x0) / a) ** 2 + ((yj - y0) / b) ** 2 < 1:
                anomaly2[i, j] = 1

    # Modify velocities and density based on anomalies
    vp[anomaly1 == 1] = np.mean(vp[anomaly1 == 1]) * 1.1
    vp[anomaly2 == 1] = np.mean(vp[anomaly2 == 1]) / 1.1

    vs[anomaly1 == 1] = np.mean(vs[anomaly1 == 1]) * 1.1
    vs[anomaly2 == 1] = np.mean(vs[anomaly2 == 1]) / 1.1

    rho[anomaly1 == 1] = np.mean(rho[anomaly1 == 1]) * 1.1
    rho[anomaly2 == 1] = np.mean(rho[anomaly2 == 1]) / 1.1
    
    # Prepare the new model with anomalies
    anomaly_model = {
        'vp': vp,
        'vs': vs,
        'rho': rho,
        'x': layer_model['x'],
        'y': layer_model['y'],
        'dx': layer_model['dx'],
        'dy': layer_model['dy']
    }
    
    return anomaly_model  # Return the anomaly model


############################################################
#                   Overthrust model
############################################################
import h5py

def download_overthrust_model(in_dir):
    if not os.path.exists(in_dir):
        os.system("wget {} -P {}".format("https://zenodo.org/records/4252588/files/overthrust_3D_true_model.h5", in_dir))  
        os.system("wget {} -P {}".format("https://zenodo.org/records/4252588/files/overthrust_3D_initial_model.h5", in_dir))  
    return

def load_overthrust_model(in_dir):
    download_overthrust_model(in_dir)
    h5_data = h5py.File(os.path.join(in_dir,"overthrust_3D_true_model.h5"))
    data_m = np.array(h5_data["m"]).astype(float)
    data_n = np.array(h5_data["n"]).astype(float)
    data_o = np.array(h5_data["o"]).astype(float)
    data_d = np.array(h5_data["d"]).astype(float)
    # slice of the 3D velocity model
    vp = np.sqrt(1/data_m[:,:,120])*1e3
    rho = pow(vp,0.25)*310
    # get the velcoty and dencity
    nx,ny = vp.shape
    overthrust_model = {}
    overthrust_model['vp'] = vp.T
    overthrust_model['rho'] = rho.T
    overthrust_model['x'] = np.arange(ny)*data_d[0]
    overthrust_model['z'] = np.arange(nx)*data_d[1]
    return overthrust_model

def load_overthrust_initial_model(in_dir):
    download_overthrust_model(in_dir)
    h5_data = h5py.File(os.path.join(in_dir,"overthrust_3D_initial_model.h5"))
    data_m = np.array(h5_data["m0"]).astype(float)
    data_n = np.array(h5_data["n"]).astype(float)
    data_o = np.array(h5_data["o"]).astype(float)
    data_d = np.array(h5_data["d"]).astype(float)
    # slice of the 3D velocity model
    vp = np.sqrt(1/data_m[:,:,120])*1e3
    rho = pow(vp,0.25)*310
    # get the velcoty and dencity
    nx,ny = vp.shape
    overthrust_model = {}
    overthrust_model['vp'] = vp.T
    overthrust_model['rho'] = rho.T
    overthrust_model['x'] = np.arange(ny)*data_d[0]
    overthrust_model['z'] = np.arange(nx)*data_d[1]
    return overthrust_model

def resample_overthrust_model(model,subsampling=2):
    vp = model["vp"]
    vp_range = vp[50:450,:200][::subsampling,::subsampling]
    rho_range = pow(vp_range,0.25)*310
    nx,ny = vp_range.shape
    overthrust_model = {}
    overthrust_model['vp']  = vp_range
    overthrust_model['rho'] = rho_range
    overthrust_model['x']   = np.arange(ny)*50
    overthrust_model['y']   = np.arange(nx)*50
    return overthrust_model


############################################################
#                   Foothill model
############################################################
# the source data is from  https://github.com/seisfwi/SWIT
def download_foothill(in_dir):
    if not os.path.exists(in_dir):
        os.system("wget {} -P {}".format("https://github.com/seisfwi/SWIT/blob/main/examples/case-07-foothill/model/Foothill_801_321_25m.dat", in_dir))
        os.system("wget {} -P {}".format("https://github.com/seisfwi/SWIT/blob/main/examples/case-07-foothill/model/Foothill_801_331_25m.dat", in_dir))
        os.system("wget {} -P {}".format("https://github.com/seisfwi/SWIT/blob/main/examples/case-07-foothill/model/Foothill_801_331_25m_smooth25.dat", in_dir))
    return

############################################################
#                   Valhall model
############################################################
def download_valhall(in_dir):
    if not os.path.exists(in_dir):
        os.system("wget {} -P {}".format("https://www.geoazur.fr/WIND/pub/nfs/FWI-DATA/GEOMODELS/Valhall2D/vp_true.bin", in_dir))
        os.system("wget {} -P {}".format("https://www.geoazur.fr/WIND/pub/nfs/FWI-DATA/GEOMODELS/Valhall2D/rho_true.bin", in_dir))
        os.system("wget {} -P {}".format("https://www.geoazur.fr/WIND/pub/nfs/FWI-DATA/GEOMODELS/Valhall2D/delta_true.bin", in_dir))
        os.system("wget {} -P {}".format("https://www.geoazur.fr/WIND/pub/nfs/FWI-DATA/GEOMODELS/Valhall2D/epsilon_true.bin", in_dir))
        os.system("wget {} -P {}".format("https://www.geoazur.fr/WIND/pub/nfs/FWI-DATA/GEOMODELS/Valhall2D/vnmo_true.bin", in_dir))
        os.system("wget {} -P {}".format("https://www.geoazur.fr/WIND/pub/nfs/FWI-DATA/GEOMODELS/Valhall2D/eta_true.bin", in_dir))
        os.system("wget {} -P {}".format("https://www.geoazur.fr/WIND/pub/nfs/FWI-DATA/GEOMODELS/Valhall2D/psimage.sh", in_dir))
    return

def load_valhall_model(in_dir):
    download_valhall(in_dir)
    
    def load_data(file_path):
        dtype = dtype = np.float32
        with open(file_path, 'rb') as f:
            data = np.fromfile(f, dtype=dtype)
        data = data.reshape(-1,209).T
        return data[3:203,:640]
    
    vp_true_path = os.path.join(in_dir,"vp_true.bin")
    rho_true_path = os.path.join(in_dir,"rho_true.bin")
    
    vp  = load_data(vp_true_path)
    rho = load_data(rho_true_path)
    dx = dz = 25
    # get the velcoty and dencity
    nx,ny = vp.shape
    valhall_model = {}
    valhall_model['vp']  = vp.T
    valhall_model['rho'] = rho.T
    valhall_model['dx'] = 25
    valhall_model['dz'] = 25
    valhall_model['x'] = np.arange(ny)*dx
    valhall_model['z'] = np.arange(nx)*dz
    return valhall_model

def get_smooth_valhall_model(model, gaussian_kernel=10):
    
    # Create copies of the velocity and density models for smoothing
    vp = model['vp'].copy()
    rho = model['rho'].copy()
    # Smooth the entire model
    vp  = gaussian_filter(vp , [gaussian_kernel, gaussian_kernel], mode='reflect')
    rho = gaussian_filter(rho, [gaussian_kernel, gaussian_kernel], mode='reflect')

    # Create a new dictionary for the smoothed model data
    new_model = {
        'vp': vp,
        'rho': rho,
        'x': model['x'],
        'z': model['z'],
        'dx': model['dx'],
        'dz': model['dz']
    }
    
    return new_model

############################################################
#                   Hess model
############################################################
import gzip
import shutil
import segyio

def download_hess(in_dir):
    if not os.path.exists(in_dir):
        os.system("wget {} -P {}".format("https://s3.amazonaws.com/open.source.geoscience/open_data/hessvti/timodel_c11.segy.gz", in_dir))
        os.system("wget {} -P {}".format("https://s3.amazonaws.com/open.source.geoscience/open_data/hessvti/timodel_c13.segy.gz", in_dir))
        os.system("wget {} -P {}".format("https://s3.amazonaws.com/open.source.geoscience/open_data/hessvti/timodel_c33.segy.gz", in_dir))
        os.system("wget {} -P {}".format("https://s3.amazonaws.com/open.source.geoscience/open_data/hessvti/timodel_c44.segy.gz", in_dir))
        os.system("wget {} -P {}".format("https://s3.amazonaws.com/open.source.geoscience/open_data/hessvti/timodel_crho.segy.gz", in_dir))
        os.system("wget {} -P {}".format("https://s3.amazonaws.com/open.source.geoscience/open_data/hessvti/timodel_epsilon.segy.gz", in_dir))
        os.system("wget {} -P {}".format("https://s3.amazonaws.com/open.source.geoscience/open_data/hessvti/timodel_delta.segy.gz", in_dir))
        os.system("wget {} -P {}".format("https://s3.amazonaws.com/open.source.geoscience/open_data/hessvti/timodel_vp.segy.gz", in_dir))
    return

def load_hess_model(in_dir):
    download_hess(in_dir)
    
    file_list = os.listdir(in_dir)
    gz_file_list = []
    for file in file_list:
        if file.endswith(".gz"):
            gz_file_list.append(file)
    
    for file in gz_file_list:
        temp_file = file.replace(".gz","")
        if not temp_file in file_list:
            segy_gz_path = os.path.join(in_dir,file)
            segy_path = os.path.join(in_dir,temp_file)
            with gzip.open(segy_gz_path, 'rb') as read, open(segy_path, 'wb') as write:
                shutil.copyfileobj(read, write)
    
    vp_path = os.path.join(in_dir,"timodel_vp.segy")
    with segyio.open(vp_path,'r',ignore_geometry=True) as f:
        vp = np.array([f.trace[i] for i in range(f.tracecount)]).T/3.33333333

    rho_path = os.path.join(in_dir,"timodel_crho.segy")
    with segyio.open(rho_path,'r',ignore_geometry=True) as f:
        rho = np.array([f.trace[i] for i in range(f.tracecount)]).T

    delta_path = os.path.join(in_dir,"timodel_delta.segy")
    with segyio.open(delta_path,'r',ignore_geometry=True) as f:
        delta = np.array([f.trace[i] for i in range(f.tracecount)]).T
    
    epsilon_path = os.path.join(in_dir,"timodel_epsilon.segy")
    with segyio.open(epsilon_path,'r',ignore_geometry=True) as f:
        epsilon = np.array([f.trace[i] for i in range(f.tracecount)]).T

    dx = dz = 25
    # get the velcoty and dencity
    nz,nx = vp.shape
    hess_model = {}
    hess_model['vp']        = vp[:1500,:3600]
    hess_model['rho']       = rho[:1500,:3600]*1000
    hess_model['delta']     = delta[:1500,:3600]
    hess_model['epsilon']   = epsilon[:1500,:3600]
    hess_model['dx'] = 10
    hess_model['dz'] = 10
    hess_model['x'] = np.arange(nx)*dx
    hess_model['z'] = np.arange(nz)*dz
    return hess_model

def get_smooth_hess_model(model, gaussian_kernel=10):
    
    # Create copies of the velocity and density models for smoothing
    vp      = model['vp'].copy()
    rho     = model['rho'].copy()
    delta   = model['delta'].copy()
    epsilon = model['epsilon'].copy()
    
    # Smooth the entire model
    vp      = gaussian_filter(vp, [gaussian_kernel, gaussian_kernel], mode='reflect')
    rho     = gaussian_filter(rho, [gaussian_kernel, gaussian_kernel], mode='reflect')
    delta   = gaussian_filter(delta, [gaussian_kernel, gaussian_kernel], mode='reflect')
    epsilon = gaussian_filter(epsilon, [gaussian_kernel, gaussian_kernel], mode='reflect')

    # Create a new dictionary for the smoothed model data
    new_model = {
        'vp': vp,
        'rho': rho,
        'delta':delta,
        'epsilon':epsilon,
        'x': model['x'],
        'z': model['z'],
        'dx': model['dx'],
        'dz': model['dz']
    }
    
    return new_model


def get_linear_hess_model(model,smooth_kernel=50,idx=None):
    """Generate a linear velocity model based on the input Marmousi model.

    Args:
        model (dict): The original Marmousi model containing velocity and density data.
        vp_min (float, optional): Minimum value for the primary wave velocity.
        vp_max (float, optional): Maximum value for the primary wave velocity.
        vs_min (float, optional): Minimum value for the shear wave velocity.
        vs_max (float, optional): Maximum value for the shear wave velocity.

    Returns:
        dict: A new model dictionary with linearly varying velocities and original density.
    """
    
    vp_true  = np.array(model['vp'])
    rho_true = np.array(model['rho'])
    delta_true = np.array(model['delta'])
    eps_true    = np.array(model['epsilon'])

    if idx is None:
        vp      = np.ones_like(vp_true)*np.mean(vp_true,axis=1).reshape(-1,1)
        rho     = np.ones_like(rho_true)*np.mean(rho_true,axis=1).reshape(-1,1)
        epsilon = np.ones_like(eps_true)*np.mean(eps_true,axis=1).reshape(-1,1)
        delta   = np.ones_like(delta_true)*np.mean(delta_true,axis=1).reshape(-1,1)
    else:
        vp      = np.ones_like(vp_true)*vp_true[:,idx].reshape(-1,1)    
        rho     = np.ones_like(rho_true)*rho_true[:,idx].reshape(-1,1)    
        epsilon = np.ones_like(eps_true)*eps_true[:,idx].reshape(-1,1)    
        delta   = np.ones_like(delta_true)*delta_true[:,idx].reshape(-1,1)    
        
    vp      = gaussian_filter1d(vp      , sigma=smooth_kernel, axis=0)
    rho     = gaussian_filter1d(rho     , sigma=smooth_kernel, axis=0)
    epsilon = gaussian_filter1d(epsilon , sigma=smooth_kernel, axis=0)
    delta   = gaussian_filter1d(delta   , sigma=smooth_kernel, axis=0)
    
    # Create a new dictionary to hold the new model data
    new_model = {
        'vp': vp,
        'rho': rho,
        'delta':delta,
        'epsilon':epsilon,
        'x': model['x'],
        'z': model['z'],
        'dx': model['dx'],
        'dz': model['dz']
    }
    
    return new_model